import logging
import time

from palimpzest.core.data.dataclasses import PlanStats
from palimpzest.core.elements.records import DataRecordCollection
from palimpzest.query.operators.aggregate import AggregateOp
from palimpzest.query.operators.filter import FilterOp
from palimpzest.query.operators.limit import LimitScanOp
from palimpzest.query.operators.scan import ScanPhysicalOp
from palimpzest.query.optimizer.plan import PhysicalPlan
from palimpzest.query.processor.query_processor import QueryProcessor
from palimpzest.sets import Dataset

logger = logging.getLogger(__name__)

class StreamingQueryProcessor(QueryProcessor):
    """This class can be used for a streaming, record-based execution.
    Results are returned as an iterable that can be consumed by the caller."""

    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)
        self._plan: PhysicalPlan | None = None
        self._plan_stats: PlanStats | None = None
        self.last_record = False
        self.current_scan_idx: int = 0
        self.plan_generated: bool = False
        self.records_count: int = 0
        logger.info("Initialized StreamingQueryProcessor")

    @property
    def plan(self) -> PhysicalPlan:
        if self._plan is None:
            raise Exception("Plan has not been generated yet.")
        return self._plan

    @plan.setter
    def plan(self, plan: PhysicalPlan):
        self._plan = plan

    @property
    def plan_stats(self) -> PlanStats:
        if self._plan_stats is None:
            raise Exception("Plan stats have not been generated yet.")
        return self._plan_stats

    @plan_stats.setter
    def plan_stats(self, plan_stats: PlanStats):
        self._plan_stats = plan_stats

    def generate_plan(self, dataset: Dataset):
        # self.clear_cached_examples()
        start_time = time.time()

        # check that the plan does not contain any aggregation operators
        for op in self.plan.operators:
            if isinstance(op, AggregateOp):
                raise Exception("You cannot have a Streaming Execution if there is an Aggregation Operator")

        # TODO: Do we need to re-initialize the optimizer here? 
        # Effectively always use the optimal strategy   
        optimizer = self.optimizer.deepcopy_clean()
        plans = optimizer.optimize(dataset)
        self.plan = plans[0]
        self.plan_stats = PlanStats.from_plan(self.plan)
        self.plan_stats.start()
        logger.info(f"Time for planning: {time.time() - start_time:.2f} seconds")
        self.plan_generated = True
        logger.info(f"Generated plan:\n{self.plan}")
        return self.plan

    def execute(self):
        logger.info("Executing StreamingQueryProcessor")
        # Always delete cache
        if not self.plan_generated:
            self.generate_plan(self.dataset)

        # if dry_run:
        #     yield [], self.plan, self.plan_stats
        #     return

        input_records = self.get_input_records()
        for idx, record in enumerate(input_records):
            # print("Iteration number: ", idx+1, "out of", len(input_records))
            output_records = self.execute_opstream(self.plan, record)
            if idx == len(input_records) - 1:
                # finalize plan stats
                self.plan_stats.finish()
            self.plan_stats.plan_str = str(self.plan)
            yield DataRecordCollection(output_records, plan_stats=self.plan_stats)

        logger.info("Done executing StreamingQueryProcessor")


    def get_input_records(self):
        scan_operator = self.plan.operators[0]
        assert isinstance(scan_operator, ScanPhysicalOp), "First operator in physical plan must be a ScanPhysicalOp"
        datareader = scan_operator.datareader
        if not datareader:
            raise Exception("DataReader not found")
        datareader_len = len(datareader)

        input_records = []
        record_op_stats = []
        for source_idx in range(datareader_len):
            record_set = scan_operator(source_idx)
            input_records += record_set.data_records
            record_op_stats += record_set.record_op_stats

        self.plan_stats.add_record_op_stats(record_op_stats)

        return input_records

    def execute_opstream(self, plan, record):
        # initialize list of output records and intermediate variables
        input_records = [record]
        record_op_stats_lst = []

        for operator in plan.operators:
            # TODO: this being defined in the for loop potentially makes the return
            # unbounded if plan.operators is empty. This should be defined outside the loop
            # and the loop refactored to account for not redeclaring this for each operator
            output_records = []

            if isinstance(operator, ScanPhysicalOp):
                continue
            # only invoke aggregate operator(s) once there are no more source records and all
            # upstream operators' processing queues are empty
            # elif isinstance(operator, AggregateOp):
            # output_records, record_op_stats_lst = operator(candidates=input_records)
            elif isinstance(operator, LimitScanOp):
                if self.records_count >= operator.limit:
                    break
            else:
                for r in input_records:
                    record_set = operator(r)
                    output_records += record_set.data_records
                    record_op_stats_lst += record_set.record_op_stats

                if isinstance(operator, FilterOp):
                    # delete all records that did not pass the filter
                    output_records = [r for r in output_records if r.passed_operator]
                    if not output_records:
                        break

            self.plan_stats.add_record_op_stats(record_op_stats_lst)
            input_records = output_records
            self.records_count += len(output_records)

        return output_records



from palimpzest.query.operators.aggregate import AggregateOp as _AggregateOp
from palimpzest.query.operators.aggregate import ApplyGroupByOp as _ApplyGroupByOp
from palimpzest.query.operators.aggregate import AverageAggregateOp as _AverageAggregateOp
from palimpzest.query.operators.aggregate import CountAggregateOp as _CountAggregateOp
from palimpzest.query.operators.convert import ConvertOp as _ConvertOp
from palimpzest.query.operators.convert import LLMConvert as _LLMConvert
from palimpzest.query.operators.convert import LLMConvertBonded as _LLMConvertBonded
from palimpzest.query.operators.convert import NonLLMConvert as _NonLLMConvert
from palimpzest.query.operators.filter import FilterOp as _FilterOp
from palimpzest.query.operators.filter import LLMFilter as _LLMFilter
from palimpzest.query.operators.filter import NonLLMFilter as _NonLLMFilter
from palimpzest.query.operators.limit import LimitScanOp as _LimitScanOp
from palimpzest.query.operators.logical import (
    Aggregate as _Aggregate,
)
from palimpzest.query.operators.logical import (
    BaseScan as _BaseScan,
)
from palimpzest.query.operators.logical import (
    CacheScan as _CacheScan,
)
from palimpzest.query.operators.logical import (
    ConvertScan as _ConvertScan,
)
from palimpzest.query.operators.logical import (
    FilteredScan as _FilteredScan,
)
from palimpzest.query.operators.logical import (
    GroupByAggregate as _GroupByAggregate,
)
from palimpzest.query.operators.logical import (
    LimitScan as _LimitScan,
)
from palimpzest.query.operators.logical import (
    LogicalOperator as _LogicalOperator,
)
from palimpzest.query.operators.logical import (
    Project as _Project,
)
from palimpzest.query.operators.logical import (
    RetrieveScan as _RetrieveScan,
)
from palimpzest.query.operators.mixture_of_agents_convert import MixtureOfAgentsConvert as _MixtureOfAgentsConvert
from palimpzest.query.operators.physical import PhysicalOperator as _PhysicalOperator
from palimpzest.query.operators.project import ProjectOp as _ProjectOp
from palimpzest.query.operators.retrieve import RetrieveOp as _RetrieveOp
from palimpzest.query.operators.scan import CacheScanDataOp as _CacheScanDataOp
from palimpzest.query.operators.scan import MarshalAndScanDataOp as _MarshalAndScanDataOp
from palimpzest.query.operators.scan import ScanPhysicalOp as _ScanPhysicalOp

LOGICAL_OPERATORS = [
    _LogicalOperator,
    _Aggregate,
    _BaseScan,
    _CacheScan,
    _ConvertScan,
    _FilteredScan,
    _GroupByAggregate,
    _LimitScan,
    _Project,
    _RetrieveScan,
]

PHYSICAL_OPERATORS = (
    # aggregate
    [_AggregateOp, _ApplyGroupByOp, _AverageAggregateOp, _CountAggregateOp]
    # convert
    + [_ConvertOp, _NonLLMConvert, _LLMConvert, _LLMConvertBonded]
    # scan
    + [_ScanPhysicalOp, _MarshalAndScanDataOp, _CacheScanDataOp]
    # filter
    + [_FilterOp, _NonLLMFilter, _LLMFilter]
    # limit
    + [_LimitScanOp]
    # mixture-of-agents
    + [_MixtureOfAgentsConvert]
    # physical
    + [_PhysicalOperator]
    # project
    + [_ProjectOp]
    # retrieve
    + [_RetrieveOp]
)

__all__ = [
    "LOGICAL_OPERATORS",
    "PHYSICAL_OPERATORS",
]



import logging

from palimpzest.core.data.dataclasses import ExecutionStats, SentinelPlanStats
from palimpzest.core.elements.records import DataRecordCollection
from palimpzest.query.optimizer.cost_model import SampleBasedCostModel
from palimpzest.query.optimizer.optimizer_strategy_type import OptimizationStrategyType
from palimpzest.query.optimizer.plan import SentinelPlan
from palimpzest.query.processor.query_processor import QueryProcessor

logger = logging.getLogger(__name__)

class SentinelQueryProcessor(QueryProcessor):

    def _generate_sample_observations(self, sentinel_plan: SentinelPlan) -> SentinelPlanStats:
        """
        This function is responsible for generating sample observation data which can be
        consumed by the CostModel.

        To accomplish this, we construct a special sentinel plan using the Optimizer which is
        capable of executing any valid physical implementation of a Filter or Convert operator
        on each record.
        """
        # if we're using validation data, get the set of expected output records
        expected_outputs = {}
        for source_idx in range(len(self.val_datasource)):
            expected_output = self.val_datasource[source_idx]
            expected_outputs[source_idx] = expected_output

        # execute sentinel plan; returns sentinel_plan_stats
        return self.sentinel_execution_strategy.execute_sentinel_plan(sentinel_plan, expected_outputs)

    def _create_sentinel_plan(self) -> SentinelPlan:
        """
        Generates and returns a SentinelPlan for the given dataset.
        """
        # create a new optimizer and update its strategy to SENTINEL
        optimizer = self.optimizer.deepcopy_clean()
        optimizer.update_strategy(OptimizationStrategyType.SENTINEL)

        # create copy of dataset, but change its data source to the validation data source
        dataset = self.dataset.copy()
        dataset._set_data_source(self.val_datasource)

        # get the sentinel plan for the given dataset
        sentinel_plans = optimizer.optimize(dataset)
        sentinel_plan = sentinel_plans[0]

        return sentinel_plan

    def execute(self) -> DataRecordCollection:
        # for now, enforce that we are using validation data; we can relax this after paper submission
        if self.val_datasource is None:
            raise Exception("Make sure you are using validation data with SentinelQueryProcessor")
        logger.info(f"Executing {self.__class__.__name__}")

        # create execution stats
        execution_stats = ExecutionStats(execution_id=self.execution_id())
        execution_stats.start()

        # create sentinel plan
        sentinel_plan = self._create_sentinel_plan()

        # generate sample execution data
        sentinel_plan_stats = self._generate_sample_observations(sentinel_plan)

        # update the execution stats to account for the work done in optimization
        execution_stats.add_plan_stats(sentinel_plan_stats)
        execution_stats.finish_optimization()

        # (re-)initialize the optimizer
        optimizer = self.optimizer.deepcopy_clean()

        # construct the CostModel with any sample execution data we've gathered
        cost_model = SampleBasedCostModel(sentinel_plan_stats, self.verbose)
        optimizer.update_cost_model(cost_model)

        # execute plan(s) according to the optimization strategy
        records, plan_stats = self._execute_best_plan(self.dataset, optimizer)

        # update the execution stats to account for the work to execute the final plan
        execution_stats.add_plan_stats(plan_stats)
        execution_stats.finish()

        # construct and return the DataRecordCollection
        result = DataRecordCollection(records, execution_stats=execution_stats)
        logger.info("Done executing SentinelQueryProcessor")

        return result



"""This file contains prompts for CriticAndRefineConvert operations."""

### CRITIQUE PROMPT AND CRITERIA ###
BASE_CRITIQUE_PROMPT = """You are a helpful assistant tasked with critiquing the output of a model based on a given prompt.
Below is the original user prompt used to generate the output:

ORIGINAL PROMPT:
<<original-prompt-placeholder>>

Here is the output generated by the model:

OUTPUT:
{original_output}

Your task is to critique the output based on the following:
{critique_criteria}

{finish_instruction}
"""

# BASE_CRITIQUE_PROMPT = """You are a helpful assistant tasked with critiquing the output of a model based on a given prompt.
# Below is the original user prompt used to generate the output:

# ORIGINAL PROMPT:
# {user_prompt}

# Here is the output generated by the model:

# OUTPUT:
# {original_output}

# Your task is to critique the output based on the following:
# {critique_criteria}

# {finish_instruction}
# """

# COT_BOOL_CRITIQUE_CRITERIA = """1. Does the output adhere to the required TRUE or FALSE format?
# 2. Is the reasoning provided logically sound and well-supported by the context?
# 3. Are there any errors in applying the filter condition to the given context?
# """

# COT_BOOL_IMAGE_CRITIQUE_CRITERIA = """1. Does the output adhere to the required TRUE or FALSE format?
# 2. Is the reasoning provided logically sound and well-supported by the image(s) and/or text?
# 3. Are there any errors in applying the filter condition to the given context?
# """

COT_QA_CRITIQUE_CRITERIA = """1. Does the JSON object adhere to the required format? Highlight any structural issues.
2. Are the values of the output fields accurate based on the provided context? If any fields are incorrect or missing, provide specific examples.
3. Are there any logical errors in reasoning used to derive the output? Provide detailed feedback on potential mistakes.
"""

COT_QA_IMAGE_CRITIQUE_CRITERIA = """1. Does the JSON object adhere to the required format?
2. Are the values of the output fields accurate based on the provided image(s) and/or text?
3. Are there any logical errors in the model's reasoning when extracting information from images and text?
"""

# COT_BO0L_CRITIQUE_FINISH_INSTRUCTION = """Finish your critique with actionable recommendations for improving the model's reasoning and answer format."""
COT_QA_CRITIQUE_FINISH_INSTRUCTION = """Finish your critique with actionable recommendations for improving the JSON object."""


### REFINEMENT PROMPT AND CRITERIA ###
BASE_REFINEMENT_PROMPT = """You are a helpful assistant tasked with refining the output of a model based on a critique.
Below is the original user prompt used to generate the output:

ORIGINAL PROMPT:
<<original-prompt-placeholder>>

Here is the original output generated by the model:

OUTPUT:
{original_output}

Here is the critique of the output:

CRITIQUE:
{critique_output}

Your task is to refine the original output to address the critique. Ensure that:
{refinement_criteria}

{finish_instruction}
"""

# BASE_REFINEMENT_PROMPT = """You are a helpful assistant tasked with refining the output of a model based on a critique.
# Below is the original user prompt used to generate the output:

# ORIGINAL PROMPT:
# {user_prompt}

# Here is the output generated by the model:

# OUTPUT:
# {original_output}

# Here is the critique of the output:

# CRITIQUE:
# {critique_output}

# Your task is to refine the original output to address the critique. Ensure that:
# {refinement_criteria}

# {finish_instruction}
# """

# COT_BOOL_REFINEMENT_CRITERIA = """1. The answer adheres to the required TRUE or FALSE format.
# 2. The reasoning is logically sound and supported by the given context.
# 3. The filter condition is correctly applied.
# """

# COT_BOOL_IMAGE_REFINEMENT_CRITERIA = """1. The answer adheres to the required TRUE or FALSE format.
# 2. The reasoning correctly considers the image(s) and/or text provided in the context.
# 3. The filter condition is properly applied.
# """

COT_QA_REFINEMENT_CRITERIA = """1. The answer adheres to the required JSON format specified in the original prompt.
2. Correctly derives all values for the output fields based on the provided context.
3. Resolves any logical errors identified in the critique.
"""

COT_QA_IMAGE_REFINEMENT_CRITERIA = """1. The answer adheres to the required JSON format specified in the original prompt.
2. Correctly derives all values for the output fields based on the provided image(s) and/or text.
3. Resolves any logical errors identified in the critique.
"""

# COT_BO0L_REFINEMENT_FINISH_INSTRUCTION = """Return the improved answer."""
COT_QA_REFINEMENT_FINISH_INSTRUCTION = """Return the refined JSON object as your final answer."""


### PROMPTS FOR MOA + CRITIC; WHICH ARE BEING DEFERRED FOR NOW ###
# COT_MOA_AGG_BASE_SYSTEM_PROMPT_CRITIQUE = """You are a helpful assistant tasked with reviewing the output of a model based on a given prompt.
# Below is the original user prompt used to generate the output:

# USER PROMPT:
# {user_prompt}

# Here is the synthesized JSON object generated by the model:

# OUTPUT:
# {original_output}

# Your task is to critique the output based on the following:
# 1. Does the JSON object adhere to the required format?
# 2. Does the synthesis appropriately combine responses from multiple models, resolving conflicts where necessary?
# 3. Are there any biases, inaccuracies, or missing information in the final output?

# Finish your critique with actionable recommendations for improving the synthesized response.
# """

# COT_MOA_AGG_BASE_SYSTEM_PROMPT_REFINEMENT = """You are a helpful assistant tasked with refining the output of a model based on a critique.
# Below is the original user prompt used to generate the output:

# USER PROMPT:
# {user_prompt}

# Here is the original synthesized JSON object generated by the model:

# ORIGINAL OUTPUT:
# {original_output}

# Here is the critique of the output:

# CRITIQUE:
# {critique_output}

# Your task is to refine the synthesized JSON object based on the critique. Ensure that:
# 1. The JSON object adheres to the required format.
# 2. The synthesis properly reconciles different model responses, making informed decisions on conflicts.
# 3. The final output is accurate, unbiased, and complete.

# Return the improved JSON object.
# """

# COT_MOA_PROPOSER_BASE_SYSTEM_PROMPT_CRITIQUE = """You are a helpful assistant tasked with reviewing the output of a model based on a given prompt.
# Below is the original user prompt used to generate the output:

# USER PROMPT:
# {user_prompt}

# Here is the response generated by the model:

# OUTPUT:
# {original_output}

# Your task is to critique the output based on the following:
# 1. Is the response well-structured and does it clearly explain each output field?
# 2. Are all claims supported by the provided context? Identify any instances of hallucination or missing evidence.
# 3. Does the response cite specific parts of the context when making claims?

# Finish your critique with actionable recommendations for improving the response.
# """

# COT_MOA_PROPOSER_BASE_SYSTEM_PROMPT_REFINEMENT = """You are a helpful assistant tasked with refining the output of a model based on a critique.
# Below is the original user prompt used to generate the output:

# USER PROMPT:
# {user_prompt}

# Here is the original response generated by the model:

# ORIGINAL OUTPUT:
# {original_output}

# Here is the critique of the output:

# CRITIQUE:
# {critique_output}

# Your task is to refine the response based on the critique. Ensure that:
# 1. The response is well-structured and clearly explains each output field.
# 2. All claims are directly supported by the provided context.
# 3. The response explicitly cites relevant parts of the context.

# Return the improved response.
# """



import logging
from copy import deepcopy
from itertools import combinations

from palimpzest.constants import AggFunc, Cardinality, PromptStrategy
from palimpzest.query.operators.aggregate import ApplyGroupByOp, AverageAggregateOp, CountAggregateOp
from palimpzest.query.operators.code_synthesis_convert import CodeSynthesisConvertSingle
from palimpzest.query.operators.convert import LLMConvertBonded, NonLLMConvert
from palimpzest.query.operators.critique_and_refine_convert import CriticAndRefineConvert
from palimpzest.query.operators.filter import LLMFilter, NonLLMFilter
from palimpzest.query.operators.limit import LimitScanOp
from palimpzest.query.operators.logical import (
    Aggregate,
    BaseScan,
    CacheScan,
    ConvertScan,
    FilteredScan,
    GroupByAggregate,
    LimitScan,
    MapScan,
    Project,
    RetrieveScan,
)
from palimpzest.query.operators.map import MapOp
from palimpzest.query.operators.mixture_of_agents_convert import MixtureOfAgentsConvert
from palimpzest.query.operators.project import ProjectOp
from palimpzest.query.operators.rag_convert import RAGConvert
from palimpzest.query.operators.retrieve import RetrieveOp
from palimpzest.query.operators.scan import CacheScanDataOp, MarshalAndScanDataOp
from palimpzest.query.operators.split_convert import SplitConvert
from palimpzest.query.optimizer.primitives import Expression, Group, LogicalExpression, PhysicalExpression
from palimpzest.utils.model_helpers import get_models, get_vision_models

logger = logging.getLogger(__name__)


class Rule:
    """
    The abstract base class for transformation and implementation rules.
    """

    @classmethod
    def get_rule_id(cls):
        return cls.__name__

    @staticmethod
    def matches_pattern(logical_expression: LogicalExpression) -> bool:
        raise NotImplementedError("Calling this method from an abstract base class.")

    @staticmethod
    def substitute(logical_expression: LogicalExpression, **kwargs) -> set[Expression]:
        raise NotImplementedError("Calling this method from an abstract base class.")


class TransformationRule(Rule):
    """
    Base class for transformation rules which convert a logical expression to another logical expression.
    The substitute method for a TransformationRule should return all new expressions and all new groups
    which are created during the substitution.
    """

    @staticmethod
    def substitute(
        logical_expression: LogicalExpression, groups: dict[int, Group], expressions: dict[int, Expression], **kwargs
    ) -> tuple[set[LogicalExpression], set[Group]]:
        """
        This function applies the transformation rule to the logical expression, which
        potentially creates new intermediate expressions and groups.

        The function returns a tuple containing:
        - the set of all new logical expressions created when applying the transformation
        - the set of all new groups created when applying the transformation
        - the next group id (after creating any new groups)
        """
        raise NotImplementedError("Calling this method from an abstract base class.")


class PushDownFilter(TransformationRule):
    """
    If this operator is a filter, push down the filter and replace it with the
    most expensive operator in the input group.
    """

    @staticmethod
    def matches_pattern(logical_expression: Expression) -> bool:
        is_match = isinstance(logical_expression.operator, FilteredScan)
        logger.debug(f"PushDownFilter matches_pattern: {is_match} for {logical_expression}")
        return is_match

    @staticmethod
    def substitute(
        logical_expression: LogicalExpression, groups: dict[int, Group], expressions: dict[int, Expression], **kwargs
    ) -> tuple[set[LogicalExpression], set[Group]]:
        logger.debug(f"Substituting PushDownFilter for {logical_expression}")

        # initialize the sets of new logical expressions and groups to be returned
        new_logical_expressions, new_groups = set(), set()

        # for each input group, if this filter does not depend on an operator
        # in that group: then swap the group with this filter
        filter_operator: FilteredScan = logical_expression.operator
        for input_group_id in logical_expression.input_group_ids:
            input_group = groups[input_group_id]

            # if the filter's dependencies aren't contained within the input group's fields,
            # then we can not push it down into this group
            if any([field not in input_group.fields for field in filter_operator.depends_on]):
                continue

            # iterate over logical expressions
            # NOTE: we previously deepcopy'ed the logical expression to avoid modifying the original;
            #       I think I've fixed this internally, but I'm leaving this NOTE as a reminder in case
            #       we see a regression / bug in the future
            for expr in input_group.logical_expressions:
                # if the expression operator is not a convert or a filter, we cannot swap
                if not (isinstance(expr.operator, (ConvertScan, FilteredScan))):
                    continue

                # if this filter depends on a field generated by the expression we're trying to swap with, we can't swap
                if any([field in expr.generated_fields for field in filter_operator.depends_on]):
                    continue

                # create new logical expression with filter pushed down to the input group's logical expression
                new_input_group_ids = deepcopy(expr.input_group_ids)
                new_input_fields = deepcopy(expr.input_fields)
                new_depends_on_field_names = deepcopy(logical_expression.depends_on_field_names)
                new_generated_fields = deepcopy(logical_expression.generated_fields)
                new_filter_expr = LogicalExpression(
                    filter_operator,
                    input_group_ids=new_input_group_ids,
                    input_fields=new_input_fields,
                    depends_on_field_names=new_depends_on_field_names,
                    generated_fields=new_generated_fields,
                    group_id=None,
                )

                # add new_filter_expr to set of new expressions
                new_logical_expressions.add(new_filter_expr)

                # get or compute the group_id and group for this new expression
                group_id, group = None, None

                # if the expression already exists, lookup the group_id and group
                if new_filter_expr.get_expr_id() in expressions:
                    group_id = expressions[new_filter_expr.get_expr_id()].group_id
                    new_filter_expr.set_group_id(group_id)
                    group = groups[group_id]

                # otherwise, lookup or create expression's group and add it to the new expressions
                else:
                    # first, compute the fields for the group
                    all_fields = {**new_input_fields, **new_generated_fields}

                    # next, compute the properties; the properties will be identical to those of the input group
                    # EXCEPT for the filters which will change as a result of our swap
                    new_group_properties = deepcopy(input_group.properties)

                    # if the expression we're swapping with is a FilteredScan,
                    # we need to remove its filter from the input group properties
                    if isinstance(expr.operator, FilteredScan):
                        filter_str = expr.operator.filter.get_filter_str()
                        new_group_properties["filters"].remove(filter_str)

                    # finally, add the pushed-down filter to the new group's properties
                    filter_str = filter_operator.filter.get_filter_str()
                    if "filters" in new_group_properties:
                        new_group_properties["filters"].add(filter_str)
                    else:
                        new_group_properties["filters"] = set([filter_str])

                    # create group for this new filter expression
                    group = Group(
                        logical_expressions=[new_filter_expr],
                        fields=all_fields,
                        properties=new_group_properties,
                    )
                    group_id = group.group_id
                    new_filter_expr.set_group_id(group_id)

                    # if the group already exists, add the expression to that group
                    if group_id in groups:
                        group = groups[group_id]
                        group.logical_expressions.add(new_filter_expr)

                    # otherwise, add this new group to groups and to the set of new groups
                    else:
                        groups[group_id] = group
                        new_groups.add(group)

                # create final new logical expression with expr's operator pulled up
                new_expr = LogicalExpression(
                    expr.operator.copy(),
                    input_group_ids=[group_id]
                    + [g_id for g_id in logical_expression.input_group_ids if g_id != input_group_id],
                    input_fields=group.fields,
                    depends_on_field_names=expr.depends_on_field_names,
                    generated_fields=expr.generated_fields,
                    group_id=logical_expression.group_id,
                )

                # add newly created expression to set of returned expressions
                new_logical_expressions.add(new_expr)

        logger.debug(f"Done substituting PushDownFilter for {logical_expression}")

        return new_logical_expressions, new_groups


class ImplementationRule(Rule):
    """
    Base class for implementation rules which convert a logical expression to a physical expression.
    """

    pass


class NonLLMConvertRule(ImplementationRule):
    """
    Substitute a logical expression for a UDF ConvertScan with a NonLLMConvert physical implementation.
    """

    @classmethod
    def matches_pattern(cls, logical_expression: LogicalExpression) -> bool:
        is_match = isinstance(logical_expression.operator, ConvertScan) and logical_expression.operator.udf is not None
        logger.debug(f"NonLLMConvertRule matches_pattern: {is_match} for {logical_expression}")
        return is_match

    @classmethod
    def substitute(cls, logical_expression: LogicalExpression, **physical_op_params) -> set[PhysicalExpression]:
        logger.debug(f"Substituting NonLLMConvertRule for {logical_expression}")

        logical_op = logical_expression.operator

        # get initial set of parameters for physical op
        op_kwargs = logical_op.get_logical_op_params()
        op_kwargs.update(
            {
                "verbose": physical_op_params["verbose"],
                "logical_op_id": logical_op.get_logical_op_id(),
                "logical_op_name": logical_op.logical_op_name(),
            }
        )

        # construct multi-expression
        op = NonLLMConvert(**op_kwargs)
        expression = PhysicalExpression(
            operator=op,
            input_group_ids=logical_expression.input_group_ids,
            input_fields=logical_expression.input_fields,
            depends_on_field_names=logical_expression.depends_on_field_names,
            generated_fields=logical_expression.generated_fields,
            group_id=logical_expression.group_id,
        )

        deduped_physical_expressions = set([expression])
        logger.debug(f"Done substituting NonLLMConvertRule for {logical_expression}")

        return deduped_physical_expressions


class LLMConvertBondedRule(ImplementationRule):
    """
    Substitute a logical expression for a ConvertScan with a bonded convert physical implementation.
    """

    @classmethod
    def matches_pattern(cls, logical_expression: LogicalExpression) -> bool:
        is_match = isinstance(logical_expression.operator, ConvertScan) and logical_expression.operator.udf is None
        logger.debug(f"LLMConvertBondedRule matches_pattern: {is_match} for {logical_expression}")
        return is_match

    @classmethod
    def substitute(cls, logical_expression: LogicalExpression, **physical_op_params) -> set[PhysicalExpression]:
        logger.debug(f"Substituting LLMConvertBondedRule for {logical_expression}")

        logical_op = logical_expression.operator

        # get initial set of parameters for physical op
        op_kwargs = logical_op.get_logical_op_params()
        op_kwargs.update(
            {
                "verbose": physical_op_params["verbose"],
                "logical_op_id": logical_op.get_logical_op_id(),
                "logical_op_name": logical_op.logical_op_name(),
            }
        )

        # identify models which can be used strictly for text or strictly for images
        vision_models = set(get_vision_models())
        text_models = set(get_models())
        pure_text_models = {model for model in text_models if model not in vision_models}
        pure_vision_models = {model for model in vision_models if model not in text_models}

        # compute attributes about this convert operation
        is_image_conversion = any(
            [
                field.is_image_field
                for field_name, field in logical_expression.input_fields.items()
                if field_name.split(".")[-1] in logical_expression.depends_on_field_names
            ]
        )
        num_image_fields = sum(
            [
                field.is_image_field
                for field_name, field in logical_expression.input_fields.items()
                if field_name.split(".")[-1] in logical_expression.depends_on_field_names
            ]
        )
        list_image_field = any(
            [
                field.is_image_field and hasattr(field, "element_type")
                for field_name, field in logical_expression.input_fields.items()
                if field_name.split(".")[-1] in logical_expression.depends_on_field_names
            ]
        )

        physical_expressions = []
        for model in physical_op_params["available_models"]:
            # skip this model if:
            # 1. this is a pure vision model and we're not doing an image conversion, or
            # 2. this is a pure text model and we're doing an image conversion, or
            # 3. this is a vision model hosted by Together (i.e. LLAMA3 vision) and there is more than one image field
            first_criteria = model in pure_vision_models and not is_image_conversion
            second_criteria = model in pure_text_models and is_image_conversion
            third_criteria = model.is_llama_model() and model.is_vision_model() and (num_image_fields > 1 or list_image_field)
            if first_criteria or second_criteria or third_criteria:
                continue

            # construct multi-expression
            op = LLMConvertBonded(
                model=model,
                prompt_strategy=PromptStrategy.COT_QA_IMAGE if is_image_conversion else PromptStrategy.COT_QA,
                **op_kwargs,
            )
            expression = PhysicalExpression(
                operator=op,
                input_group_ids=logical_expression.input_group_ids,
                input_fields=logical_expression.input_fields,
                depends_on_field_names=logical_expression.depends_on_field_names,
                generated_fields=logical_expression.generated_fields,
                group_id=logical_expression.group_id,
            )
            physical_expressions.append(expression)

        deduped_physical_expressions = set(physical_expressions)
        logger.debug(f"Done substituting LLMConvertBondedRule for {logical_expression}")

        return deduped_physical_expressions


class CodeSynthesisConvertRule(ImplementationRule):
    """
    Base rule for code synthesis convert operators; the physical convert class
    (CodeSynthesisConvertSingle) is provided by sub-class rules.

    NOTE: we provide the physical convert class(es) in their own sub-classed rules to make
    it easier to allow/disallow groups of rules at the Optimizer level.
    """

    physical_convert_class = None  # overriden by sub-classes

    @classmethod
    def matches_pattern(cls, logical_expression: LogicalExpression) -> bool:
        logical_op = logical_expression.operator
        is_image_conversion = any(
            [
                field.is_image_field
                for field_name, field in logical_expression.input_fields.items()
                if field_name.split(".")[-1] in logical_expression.depends_on_field_names
            ]
        )
        is_match = (
            isinstance(logical_op, ConvertScan)
            and not is_image_conversion
            and logical_op.cardinality != Cardinality.ONE_TO_MANY
            and logical_op.udf is None
        )
        logger.debug(f"CodeSynthesisConvertRule matches_pattern: {is_match} for {logical_expression}")
        return is_match

    @classmethod
    def substitute(cls, logical_expression: LogicalExpression, **physical_op_params) -> set[PhysicalExpression]:
        logger.debug(f"Substituting CodeSynthesisConvertRule for {logical_expression}")

        logical_op = logical_expression.operator

        # get initial set of parameters for physical op
        op_kwargs = logical_op.get_logical_op_params()
        op_kwargs.update(
            {
                "verbose": physical_op_params["verbose"],
                "logical_op_id": logical_op.get_logical_op_id(),
                "logical_op_name": logical_op.logical_op_name(),
            }
        )

        # construct multi-expression
        op = cls.physical_convert_class(
            exemplar_generation_model=physical_op_params["champion_model"],
            code_synth_model=physical_op_params["code_champion_model"],
            fallback_model=physical_op_params["fallback_model"],
            prompt_strategy=PromptStrategy.COT_QA,
            **op_kwargs,
        )
        expression = PhysicalExpression(
            operator=op,
            input_group_ids=logical_expression.input_group_ids,
            input_fields=logical_expression.input_fields,
            depends_on_field_names=logical_expression.depends_on_field_names,
            generated_fields=logical_expression.generated_fields,
            group_id=logical_expression.group_id,
        )
        deduped_physical_expressions = set([expression])
        logger.debug(f"Done substituting CodeSynthesisConvertRule for {logical_expression}")

        return deduped_physical_expressions


class CodeSynthesisConvertSingleRule(CodeSynthesisConvertRule):
    """
    Substitute a logical expression for a ConvertScan with a (single) code synthesis physical implementation.
    """

    physical_convert_class = CodeSynthesisConvertSingle


class RAGConvertRule(ImplementationRule):
    """
    Substitute a logical expression for a ConvertScan with a RAGConvert physical implementation.
    """

    num_chunks_per_fields = [1, 2, 4]
    chunk_sizes = [1000, 2000, 4000]

    @classmethod
    def matches_pattern(cls, logical_expression: LogicalExpression) -> bool:
        logical_op = logical_expression.operator
        is_image_conversion = any(
            [
                field.is_image_field
                for field_name, field in logical_expression.input_fields.items()
                if field_name.split(".")[-1] in logical_expression.depends_on_field_names
            ]
        )
        is_match = isinstance(logical_op, ConvertScan) and not is_image_conversion and logical_op.udf is None
        logger.debug(f"RAGConvertRule matches_pattern: {is_match} for {logical_expression}")
        return is_match

    @classmethod
    def substitute(cls, logical_expression: LogicalExpression, **physical_op_params) -> set[PhysicalExpression]:
        logger.debug(f"Substituting RAGConvertRule for {logical_expression}")

        logical_op = logical_expression.operator

        # get initial set of parameters for physical op
        op_kwargs = logical_op.get_logical_op_params()
        op_kwargs.update(
            {
                "verbose": physical_op_params["verbose"],
                "logical_op_id": logical_op.get_logical_op_id(),
                "logical_op_name": logical_op.logical_op_name(),
            }
        )

        # identify models which can be used strictly for text or strictly for images
        vision_models = set(get_vision_models())
        text_models = set(get_models())
        pure_vision_models = {model for model in vision_models if model not in text_models}

        physical_expressions = []
        for model in physical_op_params["available_models"]:
            # skip this model if this is a pure image model
            if model in pure_vision_models:
                continue

            for num_chunks_per_field in cls.num_chunks_per_fields:
                for chunk_size in cls.chunk_sizes:
                    # construct multi-expression
                    op = RAGConvert(
                        model=model,
                        prompt_strategy=PromptStrategy.COT_QA,
                        num_chunks_per_field=num_chunks_per_field,
                        chunk_size=chunk_size,
                        **op_kwargs,
                    )
                    expression = PhysicalExpression(
                        operator=op,
                        input_group_ids=logical_expression.input_group_ids,
                        input_fields=logical_expression.input_fields,
                        depends_on_field_names=logical_expression.depends_on_field_names,
                        generated_fields=logical_expression.generated_fields,
                        group_id=logical_expression.group_id,
                    )
                    physical_expressions.append(expression)

        logger.debug(f"Done substituting RAGConvertRule for {logical_expression}")
        deduped_physical_expressions = set(physical_expressions)

        return deduped_physical_expressions


class MixtureOfAgentsConvertRule(ImplementationRule):
    """
    Implementation rule for the MixtureOfAgentsConvert operator.
    """

    num_proposer_models = [1, 2, 3]
    temperatures = [0.0, 0.4, 0.8]

    @classmethod
    def matches_pattern(cls, logical_expression: LogicalExpression) -> bool:
        logical_op = logical_expression.operator
        is_match = isinstance(logical_op, ConvertScan) and logical_op.udf is None
        logger.debug(f"MixtureOfAgentsConvertRule matches_pattern: {is_match} for {logical_expression}")
        return is_match

    @classmethod
    def substitute(cls, logical_expression: LogicalExpression, **physical_op_params) -> set[PhysicalExpression]:
        logger.debug(f"Substituting MixtureOfAgentsConvertRule for {logical_expression}")

        logical_op = logical_expression.operator

        # get initial set of parameters for physical op
        op_kwargs: dict = logical_op.get_logical_op_params()
        op_kwargs.update(
            {
                "verbose": physical_op_params["verbose"],
                "logical_op_id": logical_op.get_logical_op_id(),
                "logical_op_name": logical_op.logical_op_name(),
            }
        )

        # identify models which can be used strictly for text or strictly for images
        vision_models = set(get_vision_models())
        text_models = set(get_models())

        # construct set of proposer models and set of aggregator models
        num_image_fields = sum(
            [
                field.is_image_field
                for field_name, field in logical_expression.input_fields.items()
                if field_name.split(".")[-1] in logical_expression.depends_on_field_names
            ]
        )
        list_image_field = any(
            [
                field.is_image_field and hasattr(field, "element_type")
                for field_name, field in logical_expression.input_fields.items()
                if field_name.split(".")[-1] in logical_expression.depends_on_field_names
            ]
        )
        proposer_model_set, is_image_conversion = text_models, False
        if num_image_fields > 1 or list_image_field:
            proposer_model_set = [model for model in vision_models if not model.is_llama_model()]
            is_image_conversion = True
        elif num_image_fields == 1:
            proposer_model_set = vision_models
            is_image_conversion = True
        aggregator_model_set = text_models

        # filter un-available models out of sets
        proposer_model_set = {model for model in proposer_model_set if model in physical_op_params["available_models"]}
        aggregator_model_set = {
            model for model in aggregator_model_set if model in physical_op_params["available_models"]
        }

        # construct MixtureOfAgentsConvert operations for various numbers of proposer models
        # and for every combination of proposer models and aggregator model
        physical_expressions = []
        for k in cls.num_proposer_models:
            for temp in cls.temperatures:
                for proposer_models in combinations(proposer_model_set, k):
                    for aggregator_model in aggregator_model_set:
                        # construct multi-expression
                        op = MixtureOfAgentsConvert(
                            proposer_models=list(proposer_models),
                            temperatures=[temp] * len(proposer_models),
                            aggregator_model=aggregator_model,
                            proposer_prompt=op_kwargs.get("prompt"),
                            proposer_prompt_strategy=PromptStrategy.COT_MOA_PROPOSER_IMAGE
                            if is_image_conversion
                            else PromptStrategy.COT_MOA_PROPOSER,
                            aggregator_prompt_strategy=PromptStrategy.COT_MOA_AGG,
                            **op_kwargs,
                        )
                        expression = PhysicalExpression(
                            operator=op,
                            input_group_ids=logical_expression.input_group_ids,
                            input_fields=logical_expression.input_fields,
                            depends_on_field_names=logical_expression.depends_on_field_names,
                            generated_fields=logical_expression.generated_fields,
                            group_id=logical_expression.group_id,
                        )
                        physical_expressions.append(expression)

        logger.debug(f"Done substituting MixtureOfAgentsConvertRule for {logical_expression}")
        deduped_physical_expressions = set(physical_expressions)

        return deduped_physical_expressions


class CriticAndRefineConvertRule(ImplementationRule):
    """
    Implementation rule for the CriticAndRefineConvert operator.
    """

    @classmethod
    def matches_pattern(cls, logical_expression: LogicalExpression) -> bool:
        logical_op = logical_expression.operator
        is_match = isinstance(logical_op, ConvertScan) and logical_op.udf is None
        logger.debug(f"CriticAndRefineConvertRule matches_pattern: {is_match} for {logical_expression}")
        return is_match

    @classmethod
    def substitute(cls, logical_expression: LogicalExpression, **physical_op_params) -> set[PhysicalExpression]:
        logger.debug(f"Substituting CriticAndRefineConvertRule for {logical_expression}")

        logical_op = logical_expression.operator

        # Get initial parameters for physical operator
        op_kwargs = logical_op.get_logical_op_params()
        op_kwargs.update(
            {
                "verbose": physical_op_params["verbose"],
                "logical_op_id": logical_op.get_logical_op_id(),
                "logical_op_name": logical_op.logical_op_name(),
            }
        )

        # identify models which can be used strictly for text or strictly for images
        vision_models = set(get_vision_models())
        text_models = set(get_models())
        pure_text_models = {model for model in text_models if model not in vision_models}
        pure_vision_models = {model for model in vision_models if model not in text_models}

        # compute attributes about this convert operation
        is_image_conversion = any(
            [
                field.is_image_field
                for field_name, field in logical_expression.input_fields.items()
                if field_name.split(".")[-1] in logical_expression.depends_on_field_names
            ]
        )
        num_image_fields = sum(
            [
                field.is_image_field
                for field_name, field in logical_expression.input_fields.items()
                if field_name.split(".")[-1] in logical_expression.depends_on_field_names
            ]
        )
        list_image_field = any(
            [
                field.is_image_field and hasattr(field, "element_type")
                for field_name, field in logical_expression.input_fields.items()
                if field_name.split(".")[-1] in logical_expression.depends_on_field_names
            ]
        )

        # identify models which can be used for this convert operation
        models = []
        for model in physical_op_params["available_models"]:
            # skip this model if:
            # 1. this is a pure vision model and we're not doing an image conversion, or
            # 2. this is a pure text model and we're doing an image conversion, or
            # 3. this is a vision model hosted by Together (i.e. LLAMA3 vision) and there is more than one image field
            first_criteria = model in pure_vision_models and not is_image_conversion
            second_criteria = model in pure_text_models and is_image_conversion
            third_criteria = model.is_llama_model() and model.is_vision_model() and (num_image_fields > 1 or list_image_field)
            if first_criteria or second_criteria or third_criteria:
                continue

            models.append(model)

        # TODO: heuristic(s) to narrow the space of critic and refine models we consider using class attributes
        # construct CriticAndRefineConvert operations for every combination of model, critic model, and refinement model
        physical_expressions = []
        for model in models:
            for critic_model in models:
                for refine_model in models:
                    # construct multi-expression
                    op = CriticAndRefineConvert(
                        model=model,
                        prompt_strategy=PromptStrategy.COT_QA_IMAGE if is_image_conversion else PromptStrategy.COT_QA,
                        critic_model=critic_model,
                        refine_model=refine_model,
                        **op_kwargs,
                    )
                    expression = PhysicalExpression(
                        operator=op,
                        input_group_ids=logical_expression.input_group_ids,
                        input_fields=logical_expression.input_fields,
                        depends_on_field_names=logical_expression.depends_on_field_names,
                        generated_fields=logical_expression.generated_fields,
                        group_id=logical_expression.group_id,
                    )
                    physical_expressions.append(expression)

        logger.debug(f"Done substituting CriticAndRefineConvertRule for {logical_expression}")
        deduped_physical_expressions = set(physical_expressions)

        return deduped_physical_expressions


class SplitConvertRule(ImplementationRule):
    """
    Substitute a logical expression for a ConvertScan with a SplitConvert physical implementation.
    """
    num_chunks = [2, 4, 6]
    min_size_to_chunk = [1000, 4000]

    @classmethod
    def matches_pattern(cls, logical_expression: LogicalExpression) -> bool:
        logical_op = logical_expression.operator
        is_image_conversion = any(
            [
                field.is_image_field
                for field_name, field in logical_expression.input_fields.items()
                if field_name.split(".")[-1] in logical_expression.depends_on_field_names
            ]
        )
        is_match = isinstance(logical_op, ConvertScan) and not is_image_conversion and logical_op.udf is None
        logger.debug(f"SplitConvertRule matches_pattern: {is_match} for {logical_expression}")
        return is_match

    @classmethod
    def substitute(cls, logical_expression: LogicalExpression, **physical_op_params) -> set[PhysicalExpression]:
        logger.debug(f"Substituting SplitConvertRule for {logical_expression}")

        logical_op = logical_expression.operator

        # get initial set of parameters for physical op
        op_kwargs = logical_op.get_logical_op_params()
        op_kwargs.update(
            {
                "verbose": physical_op_params["verbose"],
                "logical_op_id": logical_op.get_logical_op_id(),
                "logical_op_name": logical_op.logical_op_name(),
            }
        )

        # identify models which can be used strictly for text or strictly for images
        vision_models = set(get_vision_models())
        text_models = set(get_models())
        pure_vision_models = {model for model in vision_models if model not in text_models}

        physical_expressions = []
        for model in physical_op_params["available_models"]:
            # skip this model if this is a pure image model
            if model in pure_vision_models:
                continue

            for min_size_to_chunk in cls.min_size_to_chunk:
                for num_chunks in cls.num_chunks:
                    # construct multi-expression
                    op = SplitConvert(
                        model=model,
                        num_chunks=num_chunks,
                        min_size_to_chunk=min_size_to_chunk,
                        **op_kwargs,
                    )
                    expression = PhysicalExpression(
                        operator=op,
                        input_group_ids=logical_expression.input_group_ids,
                        input_fields=logical_expression.input_fields,
                        depends_on_field_names=logical_expression.depends_on_field_names,
                        generated_fields=logical_expression.generated_fields,
                        group_id=logical_expression.group_id,
                    )
                    physical_expressions.append(expression)

        logger.debug(f"Done substituting SplitConvertRule for {logical_expression}")
        deduped_physical_expressions = set(physical_expressions)

        return deduped_physical_expressions


class RetrieveRule(ImplementationRule):
    """
    Substitute a logical expression for a RetrieveScan with a Retrieve physical implementation.
    """
    k_budgets = [1, 3, 5, 10, 15, 20, 25]

    @classmethod
    def matches_pattern(cls, logical_expression: LogicalExpression) -> bool:
        is_match = isinstance(logical_expression.operator, RetrieveScan)
        logger.debug(f"RetrieveRule matches_pattern: {is_match} for {logical_expression}")
        return is_match

    @classmethod
    def substitute(cls, logical_expression: LogicalExpression, **physical_op_params) -> set[PhysicalExpression]:
        logger.debug(f"Substituting RetrieveRule for {logical_expression}")

        logical_op = logical_expression.operator

        physical_expressions = []
        ks = cls.k_budgets if logical_op.k == -1 else [logical_op.k]
        for k in ks:
            # get initial set of parameters for physical op
            op_kwargs = logical_op.get_logical_op_params()
            op_kwargs.update(
                {
                    "verbose": physical_op_params["verbose"],
                    "logical_op_id": logical_op.get_logical_op_id(),
                    "logical_op_name": logical_op.logical_op_name(),
                    "k": k,
                }
            )

            # construct multi-expression
            op = RetrieveOp(**op_kwargs)
            expression = PhysicalExpression(
                operator=op,
                input_group_ids=logical_expression.input_group_ids,
                input_fields=logical_expression.input_fields,
                depends_on_field_names=logical_expression.depends_on_field_names,
                generated_fields=logical_expression.generated_fields,
                group_id=logical_expression.group_id,
            )

            physical_expressions.append(expression)

        logger.debug(f"Done substituting RetrieveRule for {logical_expression}")
        deduped_physical_expressions = set(physical_expressions)

        return deduped_physical_expressions


class NonLLMFilterRule(ImplementationRule):
    """
    Substitute a logical expression for a FilteredScan with a non-llm filter physical implementation.
    """

    @staticmethod
    def matches_pattern(logical_expression: LogicalExpression) -> bool:
        is_match = (
            isinstance(logical_expression.operator, FilteredScan)
            and logical_expression.operator.filter.filter_fn is not None
        )
        logger.debug(f"NonLLMFilterRule matches_pattern: {is_match} for {logical_expression}")
        return is_match

    @staticmethod
    def substitute(logical_expression: LogicalExpression, **physical_op_params) -> set[PhysicalExpression]:
        logger.debug(f"Substituting NonLLMFilterRule for {logical_expression}")

        logical_op = logical_expression.operator
        op_kwargs = logical_op.get_logical_op_params()
        op_kwargs.update(
            {
                "verbose": physical_op_params["verbose"],
                "logical_op_id": logical_op.get_logical_op_id(),
                "logical_op_name": logical_op.logical_op_name(),
            }
        )
        op = NonLLMFilter(**op_kwargs)

        expression = PhysicalExpression(
            operator=op,
            input_group_ids=logical_expression.input_group_ids,
            input_fields=logical_expression.input_fields,
            depends_on_field_names=logical_expression.depends_on_field_names,
            generated_fields=logical_expression.generated_fields,
            group_id=logical_expression.group_id,
        )
        logger.debug(f"Done substituting NonLLMFilterRule for {logical_expression}")
        deduped_physical_expressions = set([expression])

        return deduped_physical_expressions


class LLMFilterRule(ImplementationRule):
    """
    Substitute a logical expression for a FilteredScan with an llm filter physical implementation.
    """

    @staticmethod
    def matches_pattern(logical_expression: LogicalExpression) -> bool:
        is_match = (
            isinstance(logical_expression.operator, FilteredScan)
            and logical_expression.operator.filter.filter_condition is not None
        )
        logger.debug(f"LLMFilterRule matches_pattern: {is_match} for {logical_expression}")
        return is_match

    @staticmethod
    def substitute(logical_expression: LogicalExpression, **physical_op_params) -> set[PhysicalExpression]:
        logger.debug(f"Substituting LLMFilterRule for {logical_expression}")

        logical_op = logical_expression.operator
        op_kwargs = logical_op.get_logical_op_params()
        op_kwargs.update(
            {
                "verbose": physical_op_params["verbose"],
                "logical_op_id": logical_op.get_logical_op_id(),
                "logical_op_name": logical_op.logical_op_name(),
            }
        )

        # identify models which can be used strictly for text or strictly for images
        vision_models = set(get_vision_models())
        text_models = set(get_models())
        pure_text_models = {model for model in text_models if model not in vision_models}
        pure_vision_models = {model for model in vision_models if model not in text_models}

        # compute attributes about this filter operation
        is_image_filter = any(
            [
                field.is_image_field
                for field_name, field in logical_expression.input_fields.items()
                if field_name.split(".")[-1] in logical_expression.depends_on_field_names
            ]
        )
        num_image_fields = sum(
            [
                field.is_image_field
                for field_name, field in logical_expression.input_fields.items()
                if field_name.split(".")[-1] in logical_expression.depends_on_field_names
            ]
        )
        list_image_field = any(
            [
                field.is_image_field and hasattr(field, "element_type")
                for field_name, field in logical_expression.input_fields.items()
                if field_name.split(".")[-1] in logical_expression.depends_on_field_names
            ]
        )

        physical_expressions = []
        for model in physical_op_params["available_models"]:
            # skip this model if:
            # 1. this is a pure vision model and we're not doing an image filter, or
            # 2. this is a pure text model and we're doing an image filter, or
            # 3. this is a vision model hosted by Together (i.e. LLAMA3 vision) and there is more than one image field
            first_criteria = model in pure_vision_models and not is_image_filter
            second_criteria = model in pure_text_models and is_image_filter
            third_criteria = model.is_llama_model() and model.is_vision_model() and (num_image_fields > 1 or list_image_field)
            if first_criteria or second_criteria or third_criteria:
                continue

            # construct multi-expression
            op = LLMFilter(
                model=model,
                prompt_strategy=PromptStrategy.COT_BOOL_IMAGE if is_image_filter else PromptStrategy.COT_BOOL,
                **op_kwargs,
            )
            expression = PhysicalExpression(
                operator=op,
                input_group_ids=logical_expression.input_group_ids,
                input_fields=logical_expression.input_fields,
                depends_on_field_names=logical_expression.depends_on_field_names,
                generated_fields=logical_expression.generated_fields,
                group_id=logical_expression.group_id,
            )
            physical_expressions.append(expression)

        logger.debug(f"Done substituting LLMFilterRule for {logical_expression}")
        deduped_physical_expressions = set(physical_expressions)

        return deduped_physical_expressions


class AggregateRule(ImplementationRule):
    """
    Substitute the logical expression for an aggregate with its physical counterpart.
    """

    @staticmethod
    def matches_pattern(logical_expression: LogicalExpression) -> bool:
        is_match = isinstance(logical_expression.operator, Aggregate)
        logger.debug(f"AggregateRule matches_pattern: {is_match} for {logical_expression}")
        return is_match

    @staticmethod
    def substitute(logical_expression: LogicalExpression, **physical_op_params) -> set[PhysicalExpression]:
        logger.debug(f"Substituting AggregateRule for {logical_expression}")

        logical_op = logical_expression.operator
        op_kwargs = logical_op.get_logical_op_params()
        op_kwargs.update(
            {
                "verbose": physical_op_params["verbose"],
                "logical_op_id": logical_op.get_logical_op_id(),
                "logical_op_name": logical_op.logical_op_name(),
            }
        )

        op = None
        if logical_op.agg_func == AggFunc.COUNT:
            op = CountAggregateOp(**op_kwargs)
        elif logical_op.agg_func == AggFunc.AVERAGE:
            op = AverageAggregateOp(**op_kwargs)
        else:
            raise Exception(f"Cannot support aggregate function: {logical_op.agg_func}")

        expression = PhysicalExpression(
            operator=op,
            input_group_ids=logical_expression.input_group_ids,
            input_fields=logical_expression.input_fields,
            depends_on_field_names=logical_expression.depends_on_field_names,
            generated_fields=logical_expression.generated_fields,
            group_id=logical_expression.group_id,
        )

        logger.debug(f"Done substituting AggregateRule for {logical_expression}")
        deduped_physical_expressions = set([expression])

        return deduped_physical_expressions


class BasicSubstitutionRule(ImplementationRule):
    """
    For logical operators with a single physical implementation, substitute the
    logical expression with its physical counterpart.
    """

    LOGICAL_OP_CLASS_TO_PHYSICAL_OP_CLASS_MAP = {
        BaseScan: MarshalAndScanDataOp,
        CacheScan: CacheScanDataOp,
        LimitScan: LimitScanOp,
        Project: ProjectOp,
        GroupByAggregate: ApplyGroupByOp,
        MapScan: MapOp,
    }

    @classmethod
    def matches_pattern(cls, logical_expression: LogicalExpression) -> bool:
        logical_op_class = logical_expression.operator.__class__
        is_match = logical_op_class in cls.LOGICAL_OP_CLASS_TO_PHYSICAL_OP_CLASS_MAP
        logger.debug(f"BasicSubstitutionRule matches_pattern: {is_match} for {logical_expression}")
        return is_match

    @classmethod
    def substitute(cls, logical_expression: LogicalExpression, **physical_op_params) -> set[PhysicalExpression]:
        logger.debug(f"Substituting BasicSubstitutionRule for {logical_expression}")

        logical_op = logical_expression.operator
        op_kwargs = logical_op.get_logical_op_params()
        op_kwargs.update(
            {
                "verbose": physical_op_params["verbose"],
                "logical_op_id": logical_op.get_logical_op_id(),
                "logical_op_name": logical_op.logical_op_name(),
            }
        )
        physical_op_class = cls.LOGICAL_OP_CLASS_TO_PHYSICAL_OP_CLASS_MAP[logical_op.__class__]
        op = physical_op_class(**op_kwargs)

        expression = PhysicalExpression(
            operator=op,
            input_group_ids=logical_expression.input_group_ids,
            input_fields=logical_expression.input_fields,
            depends_on_field_names=logical_expression.depends_on_field_names,
            generated_fields=logical_expression.generated_fields,
            group_id=logical_expression.group_id,
        )

        logger.debug(f"Done substituting BasicSubstitutionRule for {logical_expression}")
        deduped_physical_expressions = set([expression])

        return deduped_physical_expressions



from __future__ import annotations

import json
from typing import Any as TypingAny

import pandas as pd

from palimpzest.constants import MAX_ROWS
from palimpzest.core.lib.fields import (
    BooleanField,
    BytesField,
    Field,
    FloatField,
    ImageBase64Field,
    IntField,
    ListField,
    NumericField,
    StringField,
)
from palimpzest.utils.field_helpers import construct_field_type


class SchemaMetaclass(type):
    """
    This is a metaclass for our Schema class.
    """

    def __eq__(cls, other) -> bool:
        """
        Equality function for the Schema which checks that the ordered fields and class names are the same.
        """
        return cls.get_desc() == other.get_desc()


class Schema(metaclass=SchemaMetaclass):
    """
    A Schema is defined by a set of named Fields. Much of the class is implemented in the SchemaMetaclass.
    Because Schema is a MetaClass, its fields are defined similar to how they are defined in a Python dataclass.

    For example, if you wanted to define a schema for research papers, you could define a schema
    with fields representing the paper's title, publication year, and publishing journal:

    ```python
    class ResearchPaper(Schema):
        paper_title = Field(desc="The title of a scientific paper")
        paper_year = Field(desc="The year the paper was published")
        paper_journal = Field(desc="The name of the journal that published the paper")
    ```
    """

    def __init__(self, desc: str | None = None):
        self._desc = "" if desc is None else desc

    def __str__(self) -> str:
        return f"{self.__class__.__name__}(desc={self._desc})"

    # TODO: after eliminating the metaclass, this does not work
    def __eq__(self, other) -> bool:
        """
        Equality function for the Schema which checks that the ordered fields and class names are the same.
        """
        schema = self.get_desc()
        other_schema = other.get_desc()

        return schema == other_schema

    def __hash__(self) -> int:
        """Hash function for the Schema which is a simple hash of its ordered Fields and class name."""
        ordered = self.get_desc()

        return hash(ordered.encode())

    @classmethod
    def get_desc(cls) -> str:
        """Return a description of the schema"""
        fields = cls.field_names()
        d = {k: hash(getattr(cls, k)) for k in fields}
        d["__class__"] = cls.class_name()

        return json.dumps(d, sort_keys=True)

    @classmethod
    def field_names(cls, unique=False, id="") -> list[str]:
        """
        Return a list of the fields in this Schema. The `unique` argument is used to determine if the
        class name should be prefixed to the field name for unique identification. The `id` argument is
        used to provide a unique identifier for the class name.
        """
        attributes = dir(cls)
        attributes = [attr for attr in attributes if not attr.startswith("__")]
        prefix = f"{cls.__name__}.{id}." if unique else ""
        fields = [prefix + attr for attr in attributes if isinstance(getattr(cls, attr), Field)]
        return fields

    @classmethod
    def field_desc_map(cls, unique=False, id="") -> dict[str, str]:
        """
        Return a mapping from field names to their descriptions. The `unique` argument is used to determine if the
        class name should be prefixed to the field name for unique identification. The `id` argument is
        used to provide a unique identifier for the class name.
        """
        attributes = dir(cls)
        attributes = [attr for attr in attributes if not attr.startswith("__")]
        prefix = f"{cls.__name__}.{id}." if unique else ""
        field_desc_map = {
            prefix + attr: getattr(cls, attr)._desc for attr in attributes if isinstance(getattr(cls, attr), Field)
        }
        return field_desc_map

    @classmethod
    def field_map(cls, unique=False, id="") -> dict[str, Field]:
        """
        Return a mapping from field names to their field types. The `unique` argument is used to determine if the
        class name should be prefixed to the field name for unique identification. The `id` argument is used to
        provide a unique identifier for the class name.
        """
        attributes = dir(cls)
        attributes = [attr for attr in attributes if not attr.startswith("__")]
        prefix = f"{cls.__name__}.{id}." if unique else ""
        field_map = {prefix + attr: getattr(cls, attr) for attr in attributes if isinstance(getattr(cls, attr), Field)}
        return field_map

    @classmethod
    def json_schema(cls) -> dict[str, TypingAny]:
        """The JSON representation of the Schema"""
        fields = cls.field_names()

        schema = {
            "fields": {},
            "type": "object",
            "description": cls.__doc__,
        }
        for k in fields:
            if k.startswith("_"):
                continue
            v = getattr(cls, k)
            if v is None:
                continue

            schema["fields"][k] = v.json_schema()

        return schema

    @staticmethod
    def field_to_json(field_name: str, field_value: TypingAny) -> TypingAny:
        """Return a representation of the specified field which will be used in its conversion to JSON"""
        return field_value

    @classmethod
    def union(cls, other_schema: Schema, keep_duplicates: bool = False) -> Schema:
        """Return the union of this schema with the other_schema"""
        # construct the new schema name
        schema_name = cls.class_name()
        other_schema_name = other_schema.class_name()

        # construct new schema description
        new_desc = (
            f"The union of {schema_name} and {other_schema_name}\n\n"
            f"{schema_name}:\n{cls.__doc__}\n\n"
            f"{other_schema_name}:\n{other_schema.__doc__}"
        )

        # construct new lists of field names, types, and descriptions
        # NOTE: we don't need to use unique field names because they will be injected with an ID at runtime
        new_field_names, new_field_types, new_field_descs = [], [], []
        this_field_map = cls.field_map()
        for field_name, field in this_field_map.items():
            new_field_names.append(field_name)
            new_field_types.append(field)
            new_field_descs.append(field._desc)

        other_field_map = other_schema.field_map()
        for field_name, field in other_field_map.items():
            new_field_names.append(field_name)
            new_field_types.append(field)
            new_field_descs.append(field._desc)

        # rename duplicate fields if we are keeping duplicates
        if keep_duplicates:
            dup_new_field_names = []
            for left_idx, left_field_name in enumerate(new_field_names):
                # see if there's a duplicate field name
                matching_field = False
                for right_idx in range(left_idx + 1, len(new_field_names)):
                    right_field_name = new_field_names[right_idx]
                    if left_field_name == right_field_name:
                        matching_field = True
                        break

                # if theres a matching field, add them both with their schema names
                if matching_field:
                    dup_new_field_names.append(schema_name + "_" + left_field_name)
                    dup_new_field_names.append(other_schema_name + "_" + left_field_name)
                else:
                    dup_new_field_names.append(left_field_name)

            # update new_field_names
            new_field_names = dup_new_field_names

        # Generate the schema class dynamically
        attributes = {"_desc": new_desc, "__doc__": new_desc}
        for field_name, field_type, field_desc in zip(new_field_names, new_field_types, new_field_descs):
            attributes[field_name] = field_type.__class__(desc=field_desc)

        # compute the name for the new schema
        new_schema_name = f"Schema[{sorted(new_field_names)}]"

        # Create the class dynamically
        return type(new_schema_name, (Schema,), attributes)

    @classmethod
    def project(cls, project_cols: list[str]) -> Schema:
        """Return a projection of this schema with only the project_cols"""
        # construct the new schema name
        schema_name = cls.class_name()

        # construct new schema description
        new_desc = f"A projection of {schema_name} which only contains the fields {project_cols}"

        # make sure projection column names are shortened
        project_cols = [field_name.split(".")[-1] for field_name in project_cols]

        # construct new lists of field names, types, and descriptions
        # NOTE: we don't need to use unique field names because they will be injected with an ID at runtime
        new_field_names, new_field_types, new_field_descs = [], [], []
        for field_name, field in cls.field_map().items():
            if field_name in project_cols:
                new_field_names.append(field_name)
                new_field_types.append(field)
                new_field_descs.append(field._desc)

        # Generate the schema class dynamically
        attributes = {"_desc": new_desc, "__doc__": new_desc}
        for field_name, field_type, field_desc in zip(new_field_names, new_field_types, new_field_descs):
            attributes[field_name] = field_type.__class__(desc=field_desc)

        # compute the name for the new schema
        new_schema_name = f"Schema[{sorted(new_field_names)}]"

        # Create the class dynamically
        return type(new_schema_name, (Schema,), attributes)

    @staticmethod
    def from_df(df: pd.DataFrame) -> Schema:
        # get new field names, types, and descriptions
        new_field_names, new_field_types, new_field_descs = [], [], []
        for column, dtype in zip(df.columns, df.dtypes):
            column = f"column_{column}" if isinstance(column, int) else column
            field_desc = f"The {column} column from an input DataFrame"
            if dtype == "object":
                new_field_types.append(StringField(desc=field_desc))
            elif dtype == "bool":
                new_field_types.append(BooleanField(desc=field_desc))
            elif dtype == "int64":
                new_field_types.append(IntField(desc=field_desc))
            elif dtype == "float64":
                new_field_types.append(FloatField(desc=field_desc))
            else:
                new_field_types.append(Field(desc=field_desc))

            new_field_names.append(column)
            new_field_descs.append(field_desc)

        # Generate the schema class dynamically
        desc = "Schema derived from DataFrame"
        attributes = {"_desc": desc, "__doc__": desc, "__module__": Schema.__module__}
        for field_name, field_type, field_desc in zip(new_field_names, new_field_types, new_field_descs):
            attributes[field_name] = field_type.__class__(desc=field_desc)

        # compute the name for the new schema
        new_schema_name = f"Schema[{sorted(new_field_names)}]"

        # create and return the schema
        return type(new_schema_name, (Schema,), attributes)

    @classmethod
    def from_json(cls, fields: list[dict]) -> Schema:
        return cls.add_fields(fields)

    @classmethod
    def add_fields(cls, fields: list[dict]) -> Schema:
        """Add fields to the schema

        Args:
            fields: List of dictionaries, each containing 'name', 'desc', and 'type' keys

        Returns:
            A new Schema with the additional fields
        """
        assert isinstance(fields, list), "fields must be a list of dictionaries"
        for field in fields:
            assert "name" in field, "fields must contain a 'name' key"
            assert "desc" in field, "fields must contain a 'desc' key"
            assert "type" in field, "fields must contain a 'type' key"

        # build up field names, descriptions, and types
        new_field_names = [field["name"] for field in fields]
        new_field_objs = [
            construct_field_type(field["type"], desc=field["desc"])
            for field in fields
        ]

        # construct new schema
        new_desc = f"Added fields to {cls.__name__}"
        attributes = {"_desc": new_desc, "__doc__": new_desc}
        for field_name, field_obj in zip(new_field_names, new_field_objs):
            attributes[field_name] = field_obj

        new_output_schema = type(f"{cls.__name__}Extended", (Schema,), attributes)

        # return the union of this new schema with the cls
        return cls.union(new_output_schema)

    @classmethod
    def class_name(cls) -> str:
        """Return the name of this class"""
        return cls.__name__


###################################################################################
# "Core" useful Schemas. These are Schemas that almost everyone will need.
# File, TextFile, Image, PDF, etc.
###################################################################################


# First-level Schema's
class DefaultSchema(Schema):
    """Store context data."""

    value = Field(desc="The value of the input data")


class Download(Schema):
    """A download is a URL and the contents of the download."""

    url = StringField(desc="The URL of the download")
    content = BytesField(desc="The contents of the download")
    timestamp = StringField(desc="The timestamp of the download")


class File(Schema):
    """
    A File is defined by two Fields:
    - the filename (string)
    - the contents of the file (bytes)
    """

    filename = StringField(desc="The UNIX-style name of the file")
    contents = BytesField(desc="The contents of the file")

class TextFile(Schema):
    """A text file is a File that contains only text. No binary data."""
    filename = StringField(desc="The UNIX-style name of the file")
    contents = StringField(desc="The contents of the file")

class Number(Schema):
    """Just a number. Often used for aggregates"""

    value = NumericField(desc="The value of a number")


class OperatorDerivedSchema(Schema):
    """Schema defined by an operator, e.g., a join or a group by"""


class RawJSONObject(Schema):
    """A JSON object, which is a dictionary of key-value pairs."""

    json = StringField(desc="String representation of a JSON object")


list_of_strings = ListField(StringField)
list_of_lists = ListField(ListField)
class Table(Schema):
    """A Table is an object composed of a header and rows."""

    filename = StringField(desc="The name of the file the table was extracted from")
    name = StringField(desc="The name of the table")
    header = list_of_strings(desc="The header of the table")
    # TODO currently no support for nesting data records on data records
    rows = list_of_lists(desc="The rows of the table")

    def field_to_json(self, field_name: str, field_value: TypingAny) -> TypingAny:
        """Return a truncated JSON representation for `rows` and a string representation for `header`"""
        # take the first MAX_ROWS rows in the record_dict and turn them into comma separated strings
        if field_name == "rows":
            return [",".join(map(str, row)) + "\n" for row in field_value[:MAX_ROWS]]

        elif field_name == "header":
            return ",".join(field_value)

        return field_value


class URL(Schema):
    """A URL is a string that represents a web address."""

    url = StringField(desc="A URL")


class WebPage(Schema):
    """A web page is a URL and the contents of the page."""

    # url = StringField(desc="The URL of the web page")
    text = StringField(desc="The text contents of the web page")
    html = StringField(desc="The html contents of the web page")
    timestamp = StringField(desc="The timestamp of the download")
    filename = StringField(desc="The name of the file the web page was downloaded from")


# Second-level Schemas
class ImageFile(File):
    """A file that contains an image."""

    contents = ImageBase64Field(desc="The contents of the image")


class PDFFile(File):
    """A PDF file is a File that is a PDF. It has specialized fields, font information, etc."""

    # This class is currently very impoverished. It needs a lot more fields before it can correctly represent a PDF.
    text_contents = StringField(desc="The text-only contents of the PDF")


list_of_numbers = ListField(NumericField)
class XLSFile(File):
    """An XLS file is a File that contains one or more Excel spreadsheets."""

    number_sheets = NumericField(desc="The number of sheets in the Excel file")
    sheet_names = list_of_numbers(desc="The names of the sheets in the Excel file")


# Third-level Schemas
class EquationImage(ImageFile):
    """An image that contains a mathematical equation."""

    equation_text = StringField(desc="The text representation of the equation in the image")


class PlotImage(ImageFile):
    """An image that contains a plot, such as a graph or chart."""

    plot_description = StringField(desc="A description of the plot")



from __future__ import annotations

from typing import Any, Callable


#############################
# Filters that can be applied against a particular Schema
#############################
# TODO: think through a way to give filter functions fixed strings that could not be affected by a copy
#       potentially changing the address of a function; I don't think this happens today, but it's worth safeguarding against
class Filter:
    """A filter that can be applied to a Set"""

    def __init__(self, filter_condition: str | None = None, filter_fn: Callable | None = None) -> None:
        self.filter_condition = filter_condition
        self.filter_fn = filter_fn

    def serialize(self) -> dict[str, Any]:
        return {
            "filter_condition": self.filter_condition,
            "filter_fn": self.filter_fn.__name__ if self.filter_fn is not None else None,
        }

    def get_filter_str(self) -> str:
        return self.filter_condition if self.filter_condition is not None else self.filter_fn.__name__

    def __repr__(self) -> str:
        return "Filter(" + self.get_filter_str() + ")"

    def __hash__(self) -> int:
        # custom hash function
        return hash(self.filter_condition) if self.filter_condition is not None else hash(self.filter_fn.__name__)

    def __eq__(self, other) -> bool:
        # __eq__ should be defined for consistency with __hash__
        return (
            isinstance(other, Filter)
            and self.filter_condition == other.filter_condition
            and self.filter_fn == other.filter_fn
        )

    def __str__(self) -> str:
        return self.get_filter_str()



from __future__ import annotations

import math

from palimpzest.constants import (
    MODEL_CARDS,
    NAIVE_EST_NUM_INPUT_TOKENS,
    NAIVE_EST_NUM_OUTPUT_TOKENS,
    PromptStrategy,
)
from palimpzest.core.data.dataclasses import GenerationStats, OperatorCostEstimates
from palimpzest.core.elements.records import DataRecord
from palimpzest.core.lib.fields import Field, StringField
from palimpzest.query.generators.generators import generator_factory
from palimpzest.query.operators.convert import LLMConvert


class SplitConvert(LLMConvert):
    def __init__(self, num_chunks: int = 2, min_size_to_chunk: int = 1000, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.num_chunks = num_chunks
        self.min_size_to_chunk = min_size_to_chunk
        self.split_generator = generator_factory(self.model, PromptStrategy.SPLIT_PROPOSER, self.cardinality, self.verbose)
        self.split_merge_generator = generator_factory(self.model, PromptStrategy.SPLIT_MERGER, self.cardinality, self.verbose)

        # crude adjustment factor for naive estimation in no-sentinel setting
        self.naive_quality_adjustment = 0.6

    def __str__(self):
        op = super().__str__()
        op += f"    Chunk Size: {str(self.num_chunks)}\n"
        op += f"    Min Size to Chunk: {str(self.min_size_to_chunk)}\n"
        return op

    def get_id_params(self):
        id_params = super().get_id_params()
        id_params = {"num_chunks": self.num_chunks, "min_size_to_chunk": self.min_size_to_chunk, **id_params}

        return id_params

    def get_op_params(self):
        op_params = super().get_op_params()
        return {"num_chunks": self.num_chunks, "min_size_to_chunk": self.min_size_to_chunk, **op_params}

    def naive_cost_estimates(self, source_op_cost_estimates: OperatorCostEstimates) -> OperatorCostEstimates:
        """
        Update the cost per record and quality estimates produced by LLMConvert's naive estimates.
        We adjust the cost per record to account for the reduced number of input tokens following
        the retrieval of relevant chunks, and we make a crude estimate of the quality degradation
        that results from using a downsized input (although this may in fact improve quality in
        some cases).
        """
        # get naive cost estimates from LLMConvert
        naive_op_cost_estimates = super().naive_cost_estimates(source_op_cost_estimates)

        # re-compute cost per record assuming we use fewer input tokens; naively assume a single input field
        est_num_input_tokens = NAIVE_EST_NUM_INPUT_TOKENS
        est_num_output_tokens = NAIVE_EST_NUM_OUTPUT_TOKENS
        model_conversion_usd_per_record = (
            MODEL_CARDS[self.model.value]["usd_per_input_token"] * est_num_input_tokens
            + MODEL_CARDS[self.model.value]["usd_per_output_token"] * est_num_output_tokens
        )

        # set refined estimate of cost per record
        naive_op_cost_estimates.cost_per_record = model_conversion_usd_per_record
        naive_op_cost_estimates.cost_per_record_lower_bound = naive_op_cost_estimates.cost_per_record
        naive_op_cost_estimates.cost_per_record_upper_bound = naive_op_cost_estimates.cost_per_record
        naive_op_cost_estimates.quality = (naive_op_cost_estimates.quality) * self.naive_quality_adjustment
        naive_op_cost_estimates.quality_lower_bound = naive_op_cost_estimates.quality
        naive_op_cost_estimates.quality_upper_bound = naive_op_cost_estimates.quality

        return naive_op_cost_estimates

    def is_image_conversion(self) -> bool:
        """SplitConvert is currently disallowed on image conversions, so this must be False."""
        return False

    def get_text_chunks(self, text: str, num_chunks: int) -> list[str]:
        """
        Given a text string, chunk it into num_chunks substrings of roughly equal size.
        """
        chunks = []

        idx, chunk_size = 0, math.ceil(len(text) / num_chunks)
        while idx + chunk_size < len(text):
            chunks.append(text[idx : idx + chunk_size])
            idx += chunk_size

        if idx < len(text):
            chunks.append(text[idx:])

        return chunks

    def get_chunked_candidate(self, candidate: DataRecord, input_fields: list[str]) -> list[DataRecord]:
        """
        For each text field, chunk the content. If a field is smaller than the chunk size,
        simply include the full field.
        """
        # compute mapping from each field to its chunked content
        field_name_to_chunked_content = {}
        for field_name in input_fields:
            field = candidate.get_field_type(field_name)
            content = candidate[field_name]

            # do not chunk this field if it is not a string or a list of strings
            is_string_field = isinstance(field, StringField)
            is_list_string_field = hasattr(field, "element_type") and isinstance(field.element_type, StringField)
            if not (is_string_field or is_list_string_field):
                field_name_to_chunked_content[field_name] = [content]
                continue

            # if this is a list of strings, join the strings
            if is_list_string_field:
                content = "[" + ", ".join(content) + "]"

            # skip this field if its length is less than the min size to chunk
            if len(content) < self.min_size_to_chunk:
                field_name_to_chunked_content[field_name] = [content]
                continue

            # chunk the content
            field_name_to_chunked_content[field_name] = self.get_text_chunks(content, self.num_chunks)

        # compute the true number of chunks (may be 1 if all fields are not chunked)
        num_chunks = max(len(chunks) for chunks in field_name_to_chunked_content.values())

        # create the chunked canidates
        candidates = []
        for chunk_idx in range(num_chunks):
            candidate_copy = candidate.copy()
            for field_name in input_fields:
                field_chunks = field_name_to_chunked_content[field_name]
                candidate_copy[field_name] = field_chunks[chunk_idx] if len(field_chunks) > 1 else field_chunks[0]

            candidates.append(candidate_copy)

        return candidates

    def convert(self, candidate: DataRecord, fields: dict[str, Field]) -> tuple[dict[str, list], GenerationStats]:
        # get the set of input fields to use for the convert operation
        input_fields = self.get_input_fields()

        # lookup most relevant chunks for each field using embedding search
        candidate_copy = candidate.copy()
        chunked_candidates = self.get_chunked_candidate(candidate_copy, input_fields)

        # construct kwargs for generation
        gen_kwargs = {"project_cols": input_fields, "output_schema": self.output_schema}

        # generate outputs for each chunk separately
        chunk_outputs, chunk_generation_stats_lst = [], []
        for candidate in chunked_candidates:
            _, reasoning, chunk_generation_stats, _ = self.split_generator(candidate, fields, json_output=False, **gen_kwargs)
            chunk_outputs.append(reasoning)
            chunk_generation_stats_lst.append(chunk_generation_stats)

        # call the merger
        gen_kwargs = {
            "project_cols": input_fields,
            "output_schema": self.output_schema,
            "chunk_outputs": chunk_outputs,
        }
        field_answers, _, merger_gen_stats, _ = self.split_merge_generator(candidate, fields, **gen_kwargs)

        # compute the total generation stats
        generation_stats = sum(chunk_generation_stats_lst) + merger_gen_stats

        return field_answers, generation_stats



import hashlib
import io
import json
import os
import time
from typing import BinaryIO
from zipfile import ZipFile

import pandas as pd
import requests
from fastapi import status
from pypdf import PdfReader

COSMOS_ADDRESS = "https://xdd.wisc.edu/cosmos_service"


def get_md5(file_bytes: bytes) -> str:
    if not isinstance(file_bytes, bytes):
        file_bytes = file_bytes.encode()
    return hashlib.md5(file_bytes).hexdigest()


##
# Function to extract a Cosmos parquet file to Cosmos JSON
##
def cosmos_parquet_to_json(path):
    parquet_df = pd.read_parquet(path)
    parquet_json = parquet_df.to_json()
    parquet_data = json.loads(parquet_json)

    if len(parquet_data) > 0:
        parquet_data_keys = list(parquet_data.keys())
        num_data_rows = max([int(k) for k in parquet_data[parquet_data_keys[0]]])

        row_order_parquet_data = [dict() for i in range(num_data_rows + 1)]
        for field_key, row_data in parquet_data.items():
            for row_idx, datum in row_data.items():
                row_idx_num = int(row_idx)
                row_order_parquet_data[row_idx_num][field_key] = datum

        row_order_parquet_data.sort(
            key=lambda d: (
                d["page_num"],
                d["bounding_box"][0] // 500,
                d["bounding_box"][1],
            )
        )

        edits = list()
        for e1, extraction1 in enumerate(row_order_parquet_data):
            (ext1_x1, ext1_y1, ext1_x2, ext1_y2) = extraction1["bounding_box"]
            if ext1_x1 < 500:
                continue

            ext1_page_num = extraction1["page_num"]
            found_col_break = False
            insertion_index = -1
            t1 = e1
            while t1 > 0:
                extraction2 = row_order_parquet_data[t1 - 1]
                ext2_page_num = extraction2["page_num"]
                if ext1_page_num > ext2_page_num:
                    break

                (ext2_x1, ext2_y1, ext2_x2, ext2_y2) = extraction2["bounding_box"]

                if ext1_y2 <= ext2_y1:
                    ext2_xspan = ext2_x2 - ext2_x1
                    if ext2_xspan >= 800:
                        found_col_break = True
                        insertion_index = t1 - 1
                t1 -= 1
            if found_col_break:
                edits.append(
                    {
                        "del_idx": e1,
                        "ins_idx": insertion_index,
                        "val": extraction1,
                    }
                )
        for edit_dict in edits:
            del row_order_parquet_data[edit_dict["del_idx"]]
            row_order_parquet_data.insert(edit_dict["ins_idx"], edit_dict["val"])
        row_order_parquet_data.sort(key=lambda d: (d["pdf_name"]))

        name2results = dict()
        for row_data in row_order_parquet_data:
            if row_data["pdf_name"] in name2results:
                name2results[row_data["pdf_name"]].append(row_data)
            else:
                name2results[row_data["pdf_name"]] = [row_data]

        return next(iter(name2results.items()))[1]


##
# Function to extract the text 'content' attribute from the Cosmos JSON data
##
def cosmos_json_txt(cosmos_json):
    # Initialize an empty list to store the content texts
    content_texts = []

    # Iterate over each item in the JSON data
    for item in cosmos_json:
        # Extract the 'content' attribute and add it to the list
        content_texts.append(item.get("content", ""))

    return content_texts


def cosmos_client(name: str, data: BinaryIO, output_dir: str, delay=10):
    files = [
        ("pdf", (name, data, "application/pdf")),
    ]
    print(f"Sending {name} to COSMOS")
    response = requests.post(f"{COSMOS_ADDRESS}/process/", files=files)
    print(f"Received response of  {response.json()['status_endpoint']} from COSMOS: {response.status_code}")
    # get md5 of the data
    md5 = get_md5(data)

    if response.status_code == status.HTTP_202_ACCEPTED:
        callback_endpoints = response.json()

        for retry_num in range(400):
            time.sleep(delay)
            poll = requests.get(f"{callback_endpoints['status_endpoint']}")
            print(f"Polling COSMOS on retry num {retry_num + 1}")
            if poll.status_code == status.HTTP_200_OK:
                poll_results = poll.json()
                if poll_results["job_completed"]:
                    cosmos_response = requests.get(f"{callback_endpoints['result_endpoint']}")
                    if cosmos_response.status_code == status.HTTP_200_OK:
                        data = cosmos_response.content
                        with ZipFile(io.BytesIO(data)) as z:
                            output_subdir = os.path.join(
                                output_dir, f"COSMOS_{os.path.splitext(name)[0].replace(' ', '_')}_{md5}"
                            )
                            os.makedirs(output_subdir, exist_ok=True)
                            z.extractall(path=output_subdir)
                            for file in os.listdir(output_subdir):
                                if (
                                    file.endswith(".parquet")
                                    and not file.endswith("_figures.parquet")
                                    and not file.endswith("_pdfs.parquet")
                                    and not file.endswith("_tables.parquet")
                                    and not file.endswith("_sections.parquet")
                                    and not file.endswith("_equations.parquet")
                                ):
                                    print(f"Converting {file} to JSON")
                                    # if error while converting parquet to json, skip this file
                                    try:
                                        json_data = cosmos_parquet_to_json(os.path.join(output_subdir, file))
                                        with open(
                                            os.path.join(output_subdir, f"{os.path.splitext(file)[0]}.json"), "w"
                                        ) as json_file:
                                            json.dump(json_data, json_file)
                                        with open(
                                            os.path.join(output_subdir, f"{os.path.splitext(file)[0]}.txt"), "w"
                                        ) as text_file:
                                            text_file.write("\n".join(cosmos_json_txt(json_data)))
                                        # print(f"{file} : {json_data}")

                                    except Exception as e:
                                        print(f"Error while converting {file} to JSON: {e}")
                                        pass
                        return
                        # raise RuntimeError("COSMOS data doesn't include document file for annotation")

                    else:
                        raise RuntimeError(
                            f"COSMOS Result Error - STATUS CODE: {response.status_code} - {COSMOS_ADDRESS}"
                        )
                # If not, just wait until the next iteration
                else:
                    pass

        # If we reached this point, we time out
        raise TimeoutError(f"Timed out waiting for COSMOS on retry num {retry_num + 1}")

    else:
        raise RuntimeError(f"COSMOS Error - STATUS CODE: {response.status_code} - {COSMOS_ADDRESS}")


##
# Function to extract the text from a PDF file:
# 1. Check if the text file already exists in the cache, if so, read from the cache
# 2. If not, call the cosmos_client function to process the PDF file and cache the text file
##

# TODO(Jun): 1. cosmos returns 202 for me. 2. why only accept "pypdf" and "cosmos" as pdfprocessor?
def get_text_from_pdf(filename, pdf_bytes, pdfprocessor="pypdf", enable_file_cache=True, file_cache_dir="/tmp"):
    pdf_filename = filename
    file_name = os.path.basename(pdf_filename)
    file_name_without_extension = os.path.splitext(file_name)[0]
    text_file_name = f"{file_name_without_extension}.txt"

    if pdfprocessor == "pypdf":
        pdf = PdfReader(io.BytesIO(pdf_bytes))
        all_text = ""
        for page in pdf.pages:
            all_text += page.extract_text() + "\n"
        return all_text

    else:
        # Get md5 of the pdf_bytes
        md5 = get_md5(pdf_bytes)
        cached_extraction_folder = f"COSMOS_{os.path.splitext(file_name)[0].replace(' ', '_')}_{md5}"

        # Check if pz_file_cache_dir exists in the file system
        pz_file_cache_dir = os.path.join(file_cache_dir, cached_extraction_folder)
        if enable_file_cache and os.path.exists(pz_file_cache_dir):
            print(f"File {text_file_name} already exists in system tmp folder {pz_file_cache_dir}, reading from cache")
            text_file_path = os.path.join(pz_file_cache_dir, text_file_name)
            with open(text_file_path) as file:
                text_content = file.read()
                return text_content

        # Call the cosmos_client function
        print(f"Processing {file_name} through COSMOS")
        cosmos_client(file_name, pdf_bytes, file_cache_dir)
        text_file_path = os.path.join(pz_file_cache_dir, text_file_name)
        if not os.path.exists(text_file_path):
            raise FileNotFoundError(f"Text file {text_file_name} not found in {pz_file_cache_dir}/{text_file_name}")
        with open(text_file_path) as file:
            text_content = file.read()
            return text_content



from __future__ import annotations

from pathlib import Path
from typing import Callable

import pandas as pd
from chromadb.api.models.Collection import Collection

from palimpzest.constants import AggFunc, Cardinality
from palimpzest.core.data.datareaders import DataReader
from palimpzest.core.elements.filters import Filter
from palimpzest.core.elements.groupbysig import GroupBySig
from palimpzest.core.lib.schemas import Number, Schema
from palimpzest.policy import construct_policy_from_kwargs
from palimpzest.query.processor.config import QueryProcessorConfig
from palimpzest.utils.datareader_helpers import get_local_datareader
from palimpzest.utils.hash_helpers import hash_for_serialized_dict


#####################################################
#
#####################################################
class Set:
    """
    """

    def __init__(
        self,
        source: Set | DataReader,
        schema: Schema,
        desc: str | None = None,
        filter: Filter | None = None,
        udf: Callable | None = None,
        agg_func: AggFunc | None = None,
        group_by: GroupBySig | None = None,
        project_cols: list[str] | None = None,
        index: Collection | None = None,
        search_func: Callable | None = None,
        search_attr: str | None = None,
        output_attrs: list[dict] | None = None,
        k: int | None = None,  # TODO: disambiguate `k` to be something like `retrieve_k`
        limit: int | None = None,
        cardinality: Cardinality = Cardinality.ONE_TO_ONE,
        depends_on: list[str] | None = None,
        cache: bool = False,
    ):
        self._schema = schema
        self._source = source
        self._desc = desc
        self._filter = filter
        self._udf = udf
        self._agg_func = agg_func
        self._group_by = group_by
        self._project_cols = None if project_cols is None else sorted(project_cols)
        self._index = index
        self._search_func = search_func
        self._search_attr = search_attr
        self._output_attrs = output_attrs
        self._k = k
        self._limit = limit
        self._cardinality = cardinality
        self._depends_on = [] if depends_on is None else sorted(depends_on)
        self._cache = cache

    @property
    def schema(self) -> Schema:
        return self._schema

    def _set_data_source(self, source: DataReader):
        if isinstance(self._source, Set):
            self._source._set_data_source(source)
        else:
            self._source = source

    def serialize(self):
        # NOTE: I needed to remove depends_on from the serialization dictionary because
        # the optimizer changes the name of the depends_on fields to be their "full" name.
        # This created an issue with the node.universal_identifier() not being consistent
        # after changing the field to its full name.
        d = {
            "schema": self.schema.json_schema(),
            "source": self._source.serialize(),
            "desc": repr(self._desc),
            "filter": None if self._filter is None else self._filter.serialize(),
            "udf": None if self._udf is None else self._udf.__name__,
            "agg_func": None if self._agg_func is None else self._agg_func.value,
            "cardinality": self._cardinality,
            "limit": self._limit,
            "group_by": None if self._group_by is None else self._group_by.serialize(),
            "project_cols": None if self._project_cols is None else self._project_cols,
            "index": None if self._index is None else self._index.__class__.__name__,
            "search_func": None if self._search_func is None else self._search_func.__name__,
            "search_attr": self._search_attr,
            "output_attrs": None if self._output_attrs is None else str(self._output_attrs),
            "k": self._k,
        }

        return d

    def universal_identifier(self):
        """Return a unique identifier for this Set."""
        return hash_for_serialized_dict(self.serialize())

    def json_schema(self):
        """Return the JSON schema for this Set."""
        return self.schema.json_schema()


class Dataset(Set):
    """
    A Dataset is the intended abstraction for programmers to interact with when writing PZ programs.

    Users instantiate a Dataset by specifying a `source` that either points to a DataReader
    or an existing Dataset. Users can then perform computations on the Dataset in a lazy fashion
    by leveraging functions such as `filter`, `sem_filter`, `sem_add_columns`, `aggregate`, etc.
    Underneath the hood, each of these operations creates a new Dataset. As a result, the Dataset
    defines a lineage of computation.
    """

    def __init__(
        self,
        source: str | Path | list | pd.DataFrame | DataReader | Dataset,
        schema: Schema | None = None,
        *args,
        **kwargs,
    ) -> None:
        # NOTE: this function currently assumes that DataReader will always be provided with a schema;
        #       we will relax this assumption in a subsequent PR
        # convert source into a DataReader
        updated_source = get_local_datareader(source, **kwargs) if isinstance(source, (str, Path, list, pd.DataFrame)) else source

        # get the schema
        schema = updated_source.schema if schema is None else schema

        # intialize class
        super().__init__(updated_source, schema, *args, **kwargs)

    def copy(self):
        return Dataset(
            source=self._source.copy() if isinstance(self._source, Set) else self._source,
            schema=self._schema,
            desc=self._desc,
            filter=self._filter,
            udf=self._udf,
            agg_func=self._agg_func,
            group_by=self._group_by,
            project_cols=self._project_cols,
            index=self._index,
            search_func=self._search_func,
            search_attr=self._search_attr,
            output_attrs=self._output_attrs,
            k=self._k,
            limit=self._limit,
            cardinality=self._cardinality,
            depends_on=self._depends_on,
            cache=self._cache,
        )

    def filter(
        self,
        _filter: Callable,
        depends_on: str | list[str] | None = None,
    ) -> Dataset:
        """Add a user defined function as a filter to the Set. This filter will possibly restrict the items that are returned later."""
        f = None
        if callable(_filter):
            f = Filter(filter_fn=_filter)
        else:
            error_str = f"Only support callable for filter, currently got {type(_filter)}"
            if isinstance(_filter, str):
                error_str += ". Consider using sem_filter() for semantic filters."
            raise Exception(error_str)

        if isinstance(depends_on, str):
            depends_on = [depends_on]

        return Dataset(
            source=self,
            schema=self.schema,
            filter=f,
            depends_on=depends_on,
            cache=self._cache,
        )

    def sem_filter(
        self,
        _filter: str,
        depends_on: str | list[str] | None = None,
    ) -> Dataset:
        """Add a natural language description of a filter to the Set. This filter will possibly restrict the items that are returned later."""
        f = None
        if isinstance(_filter, str):
            f = Filter(_filter)
        else:
            raise Exception("sem_filter() only supports `str` input for _filter.", type(_filter))

        if isinstance(depends_on, str):
            depends_on = [depends_on]

        return Dataset(
            source=self,
            schema=self.schema,
            filter=f,
            depends_on=depends_on,
            cache=self._cache,
        )

    def sem_add_columns(self, cols: list[dict] | type[Schema],
                        cardinality: Cardinality = Cardinality.ONE_TO_ONE,
                        depends_on: str | list[str] | None = None,
                        desc: str = "Add new columns via semantic reasoning") -> Dataset:
        """
        Add new columns by specifying the column names, descriptions, and types.
        The column will be computed during the execution of the Dataset.
        Example:
            sem_add_columns(
                [{'name': 'greeting', 'desc': 'The greeting message', 'type': str},
                 {'name': 'age', 'desc': 'The age of the person', 'type': int},
                 {'name': 'full_name', 'desc': 'The name of the person', 'type': str}]
            )
        """
        if isinstance(depends_on, str):
            depends_on = [depends_on]

        new_output_schema = None
        if isinstance(cols, list):
            new_output_schema = self.schema.add_fields(cols)
        elif issubclass(cols, Schema):
            new_output_schema = self.schema.union(cols)
        else:
            raise ValueError("`cols` must be a list of dictionaries or a Schema.")

        return Dataset(
            source=self,
            schema=new_output_schema,
            udf=None,
            cardinality=cardinality,
            depends_on=depends_on,
            desc=desc,
            cache=self._cache,
        )

    def add_columns(self, udf: Callable,
                    cols: list[dict] | type[Schema],
                    cardinality: Cardinality = Cardinality.ONE_TO_ONE,
                    depends_on: str | list[str] | None = None,
                    desc: str = "Add new columns via UDF") -> Dataset:
        """
        Add new columns by specifying UDFs.

        Examples:
            add_columns(
                udf=compute_personal_greeting,
                cols=[
                    {'name': 'greeting', 'desc': 'The greeting message', 'type': str},
                    {'name': 'age', 'desc': 'The age of the person', 'type': int},
                    {'name': 'full_name', 'desc': 'The name of the person', 'type': str},
                ]
            )
        """
        if udf is None or cols is None:
            raise ValueError("`udf` and `cols` must be provided for add_columns.")

        if isinstance(depends_on, str):
            depends_on = [depends_on]

        new_output_schema = None
        if isinstance(cols, list):
            updated_cols = []
            for col_dict in cols:
                assert isinstance(col_dict, dict), "each entry in `cols` must be a dictionary"
                assert "name" in col_dict, "each type must contain a 'name' key specifying the column name"
                assert "type" in col_dict, "each type must contain a 'type' key specifying the column type"
                col_dict["desc"] = col_dict.get("desc", "New column: " + col_dict["name"])
                updated_cols.append(col_dict)
            new_output_schema = self.schema.add_fields(updated_cols)

        elif issubclass(cols, Schema):
            new_output_schema = self.schema.union(cols)

        else:
            raise ValueError("`cols` must be a list of dictionaries or a Schema.")

        return Dataset(
            source=self,
            schema=new_output_schema,
            udf=udf,
            cardinality=cardinality,
            desc=desc,
            depends_on=depends_on,
            cache=self._cache,
        )

    def map(self, udf: Callable) -> Dataset:
        """
        Apply a UDF map function.

        Examples:
            map(udf=clean_column_values)
        """
        if udf is None:
            raise ValueError("`udf` must be provided for map.")

        return Dataset(
            source=self,
            schema=self.schema,
            udf=udf,
            cache=self._cache,
        )

    def count(self) -> Dataset:
        """Apply a count aggregation to this set"""
        return Dataset(
            source=self,
            schema=Number,
            desc="Count results",
            agg_func=AggFunc.COUNT,
            cache=self._cache,
        )

    def average(self) -> Dataset:
        """Apply an average aggregation to this set"""
        return Dataset(
            source=self,
            schema=Number,
            desc="Average results",
            agg_func=AggFunc.AVERAGE,
            cache=self._cache,
        )

    def groupby(self, groupby: GroupBySig) -> Dataset:
        return Dataset(
            source=self,
            schema=groupby.output_schema(),
            desc="Group By",
            group_by=groupby,
            cache=self._cache,
        )

    def retrieve(
        self,
        index: Collection,
        search_attr: str,
        output_attrs: list[dict] | type[Schema],
        search_func: Callable | None = None,
        k: int = -1,
    ) -> Dataset:
        """
        Retrieve the top-k nearest neighbors of the value of the `search_attr` from the `index` and
        use these results to construct the `output_attrs` field(s).
        """
        new_output_schema = None
        if isinstance(output_attrs, list):
            new_output_schema = self.schema.add_fields(output_attrs)
        elif issubclass(output_attrs, Schema):
            new_output_schema = self.schema.union(output_attrs)
        else:
            raise ValueError("`cols` must be a list of dictionaries or a Schema.")

        # TODO: revisit once we can think through abstraction(s)
        # # construct the PZIndex from the user-provided index
        # index = index_factory(index)

        return Dataset(
            source=self,
            schema=new_output_schema,
            desc="Retrieve",
            index=index,
            search_func=search_func,
            search_attr=search_attr,
            output_attrs=output_attrs,
            k=k,
            cache=self._cache,
        )

    def limit(self, n: int) -> Dataset:
        """Limit the set size to no more than n rows"""
        return Dataset(
            source=self,
            schema=self.schema,
            desc="LIMIT " + str(n),
            limit=n,
            cache=self._cache,
        )

    def project(self, project_cols: list[str] | str) -> Dataset:
        """Project the Set to only include the specified columns."""
        return Dataset(
            source=self,
            schema=self.schema.project(project_cols),
            project_cols=project_cols if isinstance(project_cols, list) else [project_cols],
            cache=self._cache,
        )

    def run(self, config: QueryProcessorConfig | None = None, **kwargs):
        """Invoke the QueryProcessor to execute the query. `kwargs` will be applied to the QueryProcessorConfig."""
        # TODO: this import currently needs to be here to avoid a circular import; we should fix this in a subsequent PR
        from palimpzest.query.processor.query_processor_factory import QueryProcessorFactory

        # as syntactic sugar, we will allow some keyword arguments to parameterize our policies
        policy = construct_policy_from_kwargs(**kwargs)
        if policy is not None:
            kwargs["policy"] = policy

        return QueryProcessorFactory.create_and_run_processor(self, config, **kwargs)



from __future__ import annotations

from typing import Any

from palimpzest.constants import Cardinality, GPT_4o_MODEL_CARD, Model
from palimpzest.core.data.dataclasses import GenerationStats, OperatorCostEstimates
from palimpzest.core.elements.records import DataRecord
from palimpzest.prompts import ADVICEGEN_PROMPT, CODEGEN_PROMPT, EXAMPLE_PROMPT
from palimpzest.query.generators.generators import code_ensemble_execution, generator_factory
from palimpzest.query.operators.convert import LLMConvert, LLMConvertBonded
from palimpzest.utils.sandbox import API

# TYPE DEFINITIONS
FieldName = str
CodeName = str
Code = str
DataRecordDict = dict[str, Any]
Exemplar = tuple[DataRecordDict, DataRecordDict]
CodeEnsemble = dict[CodeName, Code]


class CodeSynthesisConvert(LLMConvert):
    def __init__(
        self,
        exemplar_generation_model: Model = Model.GPT_4o,
        code_synth_model: Model = Model.GPT_4o,
        fallback_model: Model = Model.GPT_4o_MINI,
        *args,
        **kwargs,
    ):
        kwargs["model"] = None
        super().__init__(*args, **kwargs)

        # set models
        self.exemplar_generation_model = exemplar_generation_model
        self.code_synth_model = code_synth_model
        self.fallback_model = fallback_model

        # initialize parameters
        self.field_to_code_ensemble = None
        self.exemplars = []
        self.code_synthesized = False
        self.code_champion_generator = generator_factory(
            model=self.code_synth_model,
            prompt_strategy=self.prompt_strategy,
            cardinality=Cardinality.ONE_TO_ONE,
            verbose=self.verbose,
        )
        self.field_to_code_ensemble = {}

    def __str__(self):
        op = super().__str__()
        op += f"    Code Synth Strategy: {self.__class__.__name__}\n"
        return op

    def get_id_params(self):
        id_params = super().get_id_params()
        id_params = {
            "exemplar_generation_model": self.exemplar_generation_model.value,
            "code_synth_model": self.code_synth_model.value,
            "fallback_model": self.fallback_model.value,
            **id_params,
        }

        return id_params

    def get_op_params(self):
        op_params = super().get_op_params()
        op_params = {
            "exemplar_generation_model": self.exemplar_generation_model,
            "code_synth_model": self.code_synth_model,
            "fallback_model": self.fallback_model,
            **op_params,
        }

        return op_params

    def naive_cost_estimates(self, source_op_cost_estimates: OperatorCostEstimates) -> OperatorCostEstimates:
        """
        Currently we are using GPT-4 to generate code which we can then invoke on subsequent
        inputs to this operator. To reflect this in our naive cost estimates, we assume that
        the time_per_record is low (about the time it takes to execute a cheap python function)
        and that the cost_per_record is also low (we amortize the cost of code generation across
        all records). For our quality estimate, we naively assume some degredation in quality.
        In practice, this naive quality estimate will be overwritten by the CostModel's estimate
        once it executes a few code generated examples.
        """
        naive_op_cost_estimates = super().naive_cost_estimates(source_op_cost_estimates)
        naive_op_cost_estimates.time_per_record = 1e-5
        naive_op_cost_estimates.time_per_record_lower_bound = 1e-5
        naive_op_cost_estimates.time_per_record_upper_bound = 1e-5
        naive_op_cost_estimates.cost_per_record = 1e-6  # amortize code synth cost across records
        naive_op_cost_estimates.cost_per_record_lower_bound = 1e-6
        naive_op_cost_estimates.cost_per_record_upper_bound = 1e-6
        naive_op_cost_estimates.quality = (naive_op_cost_estimates.quality) * (GPT_4o_MODEL_CARD["code"] / 100.0) * 0.7
        naive_op_cost_estimates.quality_lower_bound = naive_op_cost_estimates.quality
        naive_op_cost_estimates.quality_upper_bound = naive_op_cost_estimates.quality

        return naive_op_cost_estimates

    def _should_synthesize(
        self, exemplars: list[Exemplar], num_exemplars: int = 1, code_regenerate_frequency: int = 200, *args, **kwargs
    ) -> bool:
        """This function determines whether code synthesis should be performed based on the strategy and the number of exemplars available."""
        raise NotImplementedError("This method should be implemented in a subclass")

    def _synthesize_field_code(
        self,
        candidate: DataRecord,
        api: API,
        output_field_name: str,
        code_ensemble_num: int = 1,  # if strategy != SINGLE
        num_exemplars: int = 1,  # if strategy != EXAMPLE_ENSEMBLE
    ) -> tuple[dict[CodeName, Code], GenerationStats]:
        """This method is responsible for synthesizing the code on a per-field basis.
        Wrapping different calls to the LLM and returning a set of per-field query statistics.
        The format of the code ensemble dictionary is {code_name: code} where code_name is a string and code is a string representing the code.
        """
        raise NotImplementedError("This method should be implemented in a subclass")

    def synthesize_code_ensemble(self, fields_to_generate, candidate: DataRecord, *args, **kwargs):
        """This function is a wrapper around specific code synthesis methods
        that wraps the synthesized code per-field in a dictionary and returns the stats object.
        """
        # synthesize the per-field code ensembles
        field_to_code_ensemble = {}
        generation_stats = GenerationStats()
        for field_name in fields_to_generate:
            api = API.from_input_output_schemas(
                input_schema=self.input_schema,
                output_schema=self.output_schema,
                field_name=field_name,
                input_fields=candidate.get_field_names(),
            )

            # TODO here _synthesize_code should be called with the right parameters per-code-strategy?!
            code_ensemble, code_synth_stats = self._synthesize_field_code(candidate, api, field_name)

            # update mapping from fields to code ensemble and generation stats
            field_to_code_ensemble[field_name] = code_ensemble
            generation_stats += code_synth_stats

            if self.verbose:
                for code_name, code in code_ensemble.items():
                    print(f"CODE NAME: {code_name}")
                    print("-----------------------")
                    print(code)

        # set field_to_code_ensemble and code_synthesized to True
        return field_to_code_ensemble, generation_stats

    def _bonded_query_fallback(
        self, candidate: DataRecord
    ) -> tuple[dict[FieldName, list[Any] | None], GenerationStats]:
        fields_to_generate = self.get_fields_to_generate(candidate)
        projected_candidate = candidate.copy(include_bytes=False, project_cols=self.depends_on)

        # execute the bonded convert
        bonded_op = LLMConvertBonded(
            input_schema=self.input_schema,
            output_schema=self.output_schema,
            model=self.exemplar_generation_model,
            prompt_strategy=self.prompt_strategy,
        )
        field_answers, generation_stats = bonded_op.convert(projected_candidate, fields_to_generate)
        assert all([field in field_answers for field in fields_to_generate]), "Not all fields were generated!"

        # for the vanilla LLMConvert, we simply replace any None values with an empty list
        field_answers = {field: [] if answers is None else answers for field, answers in field_answers.items()}

        # transform the mapping from fields to answers into a (list of) DataRecord(s)
        drs, _ = self._create_data_records_from_field_answers(field_answers, candidate)

        # NOTE: this now includes bytes input fields which will show up as: `field_name = "<bytes>"`;
        #       keep an eye out for a regression in code synth performance and revert if necessary
        # update operator's set of exemplars
        exemplars = [(projected_candidate.to_dict(include_bytes=False), dr.to_dict(include_bytes=False)) for dr in drs]
        self.exemplars.extend(exemplars)

        return field_answers, generation_stats

    def is_image_conversion(self):
        """Code synthesis is disallowed on image conversions, so this must be False."""
        return False

    def convert(
        self, candidate: DataRecord, fields: list[str] | None = None
    ) -> tuple[dict[FieldName, list[Any] | None], GenerationStats]:
        # get the dictionary fields for the candidate
        candidate_dict = candidate.to_dict(include_bytes=False, project_cols=self.depends_on)

        # Check if code was already synthesized, or if we have at least one converted sample
        generation_stats = GenerationStats()
        if self._should_synthesize():
            self.field_to_code_ensemble, total_code_synth_stats = self.synthesize_code_ensemble(fields, candidate)
            self.code_synthesized = True
            generation_stats += total_code_synth_stats

        # if we have yet to synthesize code (perhaps b/c we are waiting for more exemplars),
        # use the exemplar generation model to perform the convert (and generate high-quality
        # exemplars) using a bonded query
        if not len(self.field_to_code_ensemble):
            return self._bonded_query_fallback(candidate)

        # if we have synthesized code run it on each field
        field_answers = {}
        for field_name in fields:
            # create api instance for executing python code
            api = API.from_input_output_schemas(
                input_schema=self.input_schema,
                output_schema=self.output_schema,
                field_name=field_name,
                input_fields=candidate_dict.keys(),
            )
            code_ensemble = self.field_to_code_ensemble[field_name]

            # execute the code ensemble to get the answer
            answer, _, exec_stats = code_ensemble_execution(api, code_ensemble, candidate_dict)

            # if the answer is not None, update the field_answers
            # NOTE: the answer will not be a list because code synth. is disallowed for one-to-many converts
            if answer is not None:
                generation_stats += exec_stats
                field_answers[field_name] = [answer]

            else:
                # if there is a failure, run a conventional llm convert query for the field
                if self.verbose:
                    print(f"CODEGEN FALLING BACK TO CONVENTIONAL FOR FIELD {field_name}")

                # execute the conventional llm convert
                convert_op = LLMConvertBonded(
                    input_schema=self.input_schema,
                    output_schema=self.output_schema,
                    model=self.fallback_model,
                    prompt_strategy=self.prompt_strategy,
                )
                single_field_answers, single_field_stats = convert_op.convert(candidate, [field_name])

                # include code execution time in single_field_stats
                single_field_stats.fn_call_duration_secs += exec_stats.fn_call_duration_secs

                # update generation_stats
                generation_stats += single_field_stats

                # update field answers
                # NOTE: because code synth. is disallowed for one-to-many queries, we make the first answer a singleton
                field_answers[field_name] = (
                    [single_field_answers[field_name][0]]
                    if single_field_answers[field_name] is not None and len(single_field_answers[field_name]) > 0
                    else []
                )

        assert all([field in field_answers for field in fields]), "Not all fields were generated!"

        # for the vanilla LLMConvert, we simply replace any None values with an empty list
        field_answers = {field: [] if answers is None else answers for field, answers in field_answers.items()}

        return field_answers, generation_stats


class CodeSynthesisConvertNone(CodeSynthesisConvert):
    def _should_synthesize(self, *args, **kwargs):
        return False

    def _synthesize_field_code(self, candidate: DataRecord, api: API, *args, **kwargs):
        code = api.api_def() + "  return None\n"
        code_ensemble = {"{api.name}_v0": code}
        return code_ensemble, GenerationStats()


class CodeSynthesisConvertSingle(CodeSynthesisConvert):
    def _should_synthesize(self, num_exemplars: int = 1, *args, **kwargs) -> bool:
        """This function determines whether code synthesis
        should be performed based on the strategy and the number of exemplars available."""
        if len(self.exemplars) < num_exemplars:
            return False
        return not self.code_synthesized

    def _code_synth_single(
        self,
        candidate: DataRecord,
        api: API,
        output_field_name: str,
        exemplars: list[Exemplar] | None = None,
        advice: str | None = None,
        language="Python",
    ):
        if exemplars is None:
            exemplars = []

        context = {
            "language": language,
            "api": api.args_call(),
            "output": api.output,
            "inputs_desc": "\n".join(
                [f"- {field_name} ({api.input_descs[i]})" for i, field_name in enumerate(api.inputs)]
            ),
            "output_desc": api.output_desc,
            "examples_desc": "\n".join(
                [
                    EXAMPLE_PROMPT.format(
                        idx=f" {i}",
                        example_inputs="\n".join(
                            [f"- {field_name} = {repr(example[0][field_name])}" for field_name in example[0]]
                        ),
                        example_output=f"{example[1][output_field_name]}",
                    )
                    for i, example in enumerate(exemplars)
                ]
            ),
            "advice": f"Hint: {advice}" if advice else "",
        }

        prompt = CODEGEN_PROMPT.format(**context)
        if self.verbose:
            print("PROMPT")
            print("-------")
            print(f"{prompt}")

        # set prompt for generator
        gen_kwargs = {"prompt": prompt, "parse_answer": lambda text: text.split("answer:")[-1].split("---")[0].strip()}

        # invoke the champion model to generate the code
        pred, _, stats, _ = self.code_champion_generator(candidate, None, json_output=False, **gen_kwargs)
        ordered_keys = [f"```{language}", f"```{language.lower()}", "```"]
        code = None
        if not pred:
            return code, stats

        for key in ordered_keys:
            if key in pred:
                code = pred.split(key)[1].split("```")[0].strip()
                break

        if self.verbose:
            print("-------")
            print("SYNTHESIZED CODE")
            print("---------------")
            print(f"{code}")

        return code, stats

    def _synthesize_field_code(
        self, candidate: DataRecord, api: API, output_field_name: str, num_exemplars: int = 1, *args, **kwargs
    ):
        code, generation_stats = self._code_synth_single(
            candidate, api, output_field_name, exemplars=self.exemplars[:num_exemplars]
        )
        code_ensemble = {f"{api.name}_v0": code}
        return code_ensemble, generation_stats


# NOTE A nicer truly class based approach would re-implement the code_synth_single method with calls to
# __super__ and then only re-implement the differences instead of having the code in the superclass know
# about the subclass-specific parameters (i.e., advice).
class CodeSynthesisConvertExampleEnsemble(CodeSynthesisConvertSingle):
    def _should_synthesize(self, num_exemplars: int = 1, *args, **kwargs) -> bool:
        if len(self.exemplars) < num_exemplars:
            return False
        return not self.code_synthesized

    def _synthesize_field_code(
        self, candidate: DataRecord, api: API, output_field_name: str, code_ensemble_num: int = 1, *args, **kwargs
    ):
        # creates an ensemble of `code_ensemble_num` synthesized functions; each of
        # which uses a different exemplar (modulo the # of exemplars) for its synthesis
        code_ensemble = {}
        generation_stats = GenerationStats()
        for i in range(code_ensemble_num):
            code_name = f"{api.name}_v{i}"
            exemplar = self.exemplars[i % len(self.exemplars)]
            code, stats = self._code_synth_single(candidate, api, output_field_name, exemplars=[exemplar])
            code_ensemble[code_name] = code
            generation_stats += stats

        return code_ensemble, generation_stats


class CodeSynthesisConvertAdviceEnsemble(CodeSynthesisConvertSingle):
    def _should_synthesize(self, *args, **kwargs):
        return False

    def _parse_multiple_outputs(self, text, outputs=None):
        if outputs is None:
            outputs = ["Thought", "Action"]
        data = {}
        for key in reversed(outputs):
            if key + ":" in text:
                remain, value = text.rsplit(key + ":", 1)
                data[key.lower()] = value.strip()
                text = remain
            else:
                data[key.lower()] = None
        return data

    def _synthesize_advice(
        self,
        candidate: DataRecord,
        api: API,
        output_field_name: str,
        exemplars: list[Exemplar] | None = None,
        language="Python",
        n_advices=4,
        limit: int = 3,
    ):
        if exemplars is None:
            exemplars = []
        context = {
            "language": language,
            "api": api.args_call(),
            "output": api.output,
            "inputs_desc": "\n".join(
                [f"- {field_name} ({api.input_descs[i]})" for i, field_name in enumerate(api.inputs)]
            ),
            "output_desc": api.output_desc,
            "examples_desc": "\n".join(
                [
                    EXAMPLE_PROMPT.format(
                        idx=f" {i}",
                        example_inputs="\n".join(
                            [f"- {field_name} = {repr(example[0][field_name])}" for field_name in example[0]]
                        ),
                        example_output=f"{example[1][output_field_name]}",
                    )
                    for i, example in enumerate(exemplars)
                ]
            ),
            "n": n_advices,
        }
        prompt = ADVICEGEN_PROMPT.format(**context)

        # set prompt for generator
        gen_kwargs = {"prompt": prompt, "parse_answer": lambda text: text.split("answer:")[-1].split("---")[0].strip()}

        pred, _, stats, _ = self.code_champion_generator(candidate, None, json_output=False, **gen_kwargs)
        advs = self._parse_multiple_outputs(pred, outputs=[f"Idea {i}" for i in range(1, limit + 1)])

        return advs, stats

    def _synthesize_field_code(
        self,
        candidate: DataRecord,
        api: API,
        output_field_name: str,
        code_ensemble_num: int = 1,
        num_exemplars: int = 1,
        *args,
        **kwargs,
    ):
        # a more advanced approach in which advice is first solicited, and then
        # provided as context when synthesizing the code ensemble
        output_stats = {}
        # solicit advice
        advices, adv_stats = self._synthesize_advice(
            candidate, api, output_field_name, exemplars=self.exemplars[:num_exemplars], n_advices=code_ensemble_num
        )
        for key, value in adv_stats.items():
            if isinstance(value, dict):
                for k2, v2 in value.items():
                    output_stats[k2] = output_stats.get(k2, 0) + v2
            else:
                output_stats[key] += output_stats.get(key, type(value)()) + value

        code_ensemble = {}
        # synthesize code ensemble
        for i, adv in enumerate(advices):
            code_name = f"{api.name}_v{i}"
            code, stats = self._code_synth_single(
                candidate, api, output_field_name, exemplars=self.exemplars[:num_exemplars], advice=adv
            )
            code_ensemble[code_name] = code
            for key in output_stats:
                output_stats[key] += stats[key]
        return code_ensemble, output_stats


class CodeSynthesisConvertAdviceEnsembleValidation(CodeSynthesisConvert):
    def _should_synthesize(self, code_regenerate_frequency: int = 200, *args, **kwargs):
        return len(self.exemplars) % code_regenerate_frequency == 0

    def _synthesize_field_code(
        self, api: API, output_field_name: str, exemplars: list[Exemplar] = None, *args, **kwargs
    ):
        # TODO this was not implemented ?
        if exemplars is None:
            exemplars = []
        raise Exception("not implemented yet")



import os
import sys


def load_env():
    sys.path.insert(0, os.path.abspath(os.path.join(os.path.dirname(__file__), "..")))

    # read the env file
    if os.path.exists(".env"):
        with open(".env") as f:
            for line in f:
                key, value = line.strip().split("=")
                os.environ[key] = value



import json
from dataclasses import dataclass, field

from palimpzest.constants import Model
from palimpzest.core.data.datareaders import DataReader
from palimpzest.policy import MaxQuality, Policy


# TODO: Separate out the config for the Optimizer, ExecutionStrategy, and QueryProcessor
# TODO: Add description for each field.
@dataclass
class QueryProcessorConfig:
    """Shared context for query processors"""
    processing_strategy: str = field(default="auto")                 # substituted with ProcessingStrategyType
    execution_strategy: str = field(default="sequential")            # substituted with ExecutionStrategyType
    sentinel_execution_strategy: str | None = field(default="auto")  # substituted with SentinelExecutionStrategyType
    optimizer_strategy: str = field(default="pareto")                # substituted with OptimizationStrategyType

    val_datasource: DataReader | None = field(default=None)

    policy: Policy = field(default_factory=MaxQuality)
    scan_start_idx: int = field(default=0)
    num_samples: int = field(default=None)
    cache: bool = field(default=False)  # NOTE: until we properly implement caching, let's set the default to False
    verbose: bool = field(default=False)
    progress: bool = field(default=True)
    available_models: list[Model] | None = field(default=None)

    max_workers: int | None = field(default=None)

    allow_bonded_query: bool = field(default=True)
    allow_model_selection: bool = field(default=True)
    allow_code_synth: bool = field(default=False)
    allow_rag_reduction: bool = field(default=True)
    allow_mixtures: bool = field(default=True)
    allow_critic: bool = field(default=True)
    allow_split_merge: bool = field(default=False)
    use_final_op_quality: bool = field(default=False)

    kwargs: dict = field(default_factory=dict)

    def to_dict(self) -> dict:
        """Convert the config to a dict representation."""
        return {
            "processing_strategy": self.processing_strategy,
            "execution_strategy": self.execution_strategy,
            "sentinel_execution_strategy": self.sentinel_execution_strategy,
            "optimizer_strategy": self.optimizer_strategy,
            "val_datasource": self.val_datasource,
            "policy": self.policy,
            "scan_start_idx": self.scan_start_idx,
            "num_samples": self.num_samples,
            "cache": self.cache,
            "verbose": self.verbose,
            "progress": self.progress,
            "available_models": self.available_models,
            "max_workers": self.max_workers,
            "allow_bonded_query": self.allow_bonded_query,
            "allow_model_selection": self.allow_model_selection,
            "allow_code_synth": self.allow_code_synth,
            "allow_rag_reduction": self.allow_rag_reduction,
            "allow_mixtures": self.allow_mixtures,
            "allow_critic": self.allow_critic,
            "allow_split_merge": self.allow_split_merge,
            "use_final_op_quality": self.use_final_op_quality,
            **self.kwargs,
        }

    def to_json_str(self):
        """Convert the config to a JSON string representation."""
        config_dict = self.to_dict()
        config_dict["val_datasource"] = (
            None if self.val_datasource is None else self.val_datasource.serialize()
        )
        config_dict["policy"] = self.policy.to_json_str()
        for strategy in ["processing_strategy", "execution_strategy", "sentinel_execution_strategy", "optimizer_strategy"]:
            config_dict[strategy] = str(config_dict[strategy])

        return json.dumps(config_dict, indent=2)

    def update(self, **kwargs) -> None:
        for key, value in kwargs.items():
            if hasattr(self, key):
                setattr(self, key, value)
        
        self.kwargs.update(kwargs)



from __future__ import annotations

import time
from abc import ABC, abstractmethod

from palimpzest.constants import (
    LOCAL_SCAN_TIME_PER_KB,
    MEMORY_SCAN_TIME_PER_KB,
    Cardinality,
)
from palimpzest.core.data.dataclasses import OperatorCostEstimates, RecordOpStats
from palimpzest.core.data.datareaders import DataReader, DirectoryReader, FileReader
from palimpzest.core.elements.records import DataRecord, DataRecordSet
from palimpzest.query.operators.physical import PhysicalOperator


class ScanPhysicalOp(PhysicalOperator, ABC):
    """
    Physical operators which implement DataReaders require slightly more information
    in order to accurately compute naive cost estimates. Thus, we use a slightly
    modified abstract base class for these operators.
    """

    def __init__(self, datareader: DataReader, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.datareader = datareader

    def __str__(self):
        op = f"{self.op_name()}({self.datareader}) -> {self.output_schema}\n"
        op += f"    ({', '.join(self.output_schema.field_names())[:30]})\n"
        return op

    def get_id_params(self):
        return super().get_id_params()

    def get_op_params(self):
        op_params = super().get_op_params()
        return {"datareader": self.datareader, **op_params}

    @abstractmethod
    def naive_cost_estimates(
        self,
        source_op_cost_estimates: OperatorCostEstimates,
        input_cardinality: Cardinality,
        input_record_size_in_bytes: int | float,
    ) -> OperatorCostEstimates:
        """
        This function returns a naive estimate of this operator's:
        - cardinality
        - time_per_record
        - cost_per_record
        - quality

        For the implemented operator. These will be used by the CostModel
        when PZ does not have sample execution data -- and it will be necessary
        in some cases even when sample execution data is present. (For example,
        the cardinality of each operator cannot be estimated based on sample
        execution data alone -- thus ScanPhysicalOps need to give
        at least ballpark correct estimates of this quantity).
        """
        pass

    def __call__(self, idx: int) -> DataRecordSet:
        """
        This function invokes `self.datareader.__getitem__` on the given `idx` to retrieve the next data item.
        It then returns this item as a DataRecord wrapped in a DataRecordSet.
        """
        start_time = time.time()
        item = self.datareader[idx]
        end_time = time.time()

        # TODO: remove once validation data is refactored
        item_field_dict = item.get("fields", item)

        # check that item covers fields in output schema
        output_field_names = self.output_schema.field_names()
        assert all([field in item_field_dict for field in output_field_names]), f"Some fields in DataReader schema not present in item!\n - DataReader fields: {output_field_names}\n - Item fields: {list(item.keys())}"

        # construct a DataRecord from the item
        dr = DataRecord(self.output_schema, source_idx=idx)
        for field in output_field_names:
            setattr(dr, field, item_field_dict[field])

        # create RecordOpStats objects
        record_op_stats = RecordOpStats(
            record_id=dr.id,
            record_parent_id=dr.parent_id,
            record_source_idx=dr.source_idx,
            record_state=dr.to_dict(include_bytes=False),
            full_op_id=self.get_full_op_id(),
            logical_op_id=self.logical_op_id,
            op_name=self.op_name(),
            time_per_record=(end_time - start_time),
            cost_per_record=0.0,
            op_details={k: str(v) for k, v in self.get_id_params().items()},
        )
 
        # construct and return DataRecordSet object
        return DataRecordSet([dr], [record_op_stats])
        

class MarshalAndScanDataOp(ScanPhysicalOp):
    def naive_cost_estimates(
        self,
        source_op_cost_estimates: OperatorCostEstimates,
        input_record_size_in_bytes: int | float,
    ) -> OperatorCostEstimates:
        # get inputs needed for naive cost estimation
        # TODO: we should rename cardinality --> "multiplier" or "selectivity" one-to-one / one-to-many

        # estimate time spent reading each record
        per_record_size_kb = input_record_size_in_bytes / 1024.0
        time_per_record = (
            LOCAL_SCAN_TIME_PER_KB * per_record_size_kb
            if isinstance(self.datareader, (DirectoryReader, FileReader))
            else MEMORY_SCAN_TIME_PER_KB * per_record_size_kb
        )

        # estimate output cardinality
        cardinality = source_op_cost_estimates.cardinality

        # for now, assume no cost per record for reading data
        return OperatorCostEstimates(
            cardinality=cardinality,
            time_per_record=time_per_record,
            cost_per_record=0,
            quality=1.0,
        )


class CacheScanDataOp(ScanPhysicalOp):
    def naive_cost_estimates(
        self,
        source_op_cost_estimates: OperatorCostEstimates,
        input_record_size_in_bytes: int | float,
    ):
        # get inputs needed for naive cost estimation
        # TODO: we should rename cardinality --> "multiplier" or "selectivity" one-to-one / one-to-many

        # estimate time spent reading each record
        per_record_size_kb = input_record_size_in_bytes / 1024.0
        time_per_record = LOCAL_SCAN_TIME_PER_KB * per_record_size_kb

        # estimate output cardinality
        cardinality = source_op_cost_estimates.cardinality

        # for now, assume no cost per record for reading from cache
        return OperatorCostEstimates(
            cardinality=cardinality,
            time_per_record=time_per_record,
            cost_per_record=0,
            quality=1.0,
        )
